class SharedConstants:
    # Must be kept up to date with setup.py
    _CONVERTER21_NAME = 'converter21'
    _CONVERTER21_VERSION = '4.0.0'
    _CONVERTER21_NAME_AND_VERSION = _CONVERTER21_NAME + ' v' + _CONVERTER21_VERSION

    SMUFL_NAME_TO_UNICODE_CHAR: dict[str, str] = {
        'space': chr(0x0020),
        'musicFlatSign': chr(0x266D),
        'musicNaturalSign': chr(0x266E),
        'musicSharpSign': chr(0x266F),
        'brace': chr(0xE000),
        'bracketTop': chr(0xE003),
        'bracketBottom': chr(0xE004),
        'repeatDots': chr(0xE043),
        'repeatDot': chr(0xE044),
        'dalSegno': chr(0xE045),
        'daCapo': chr(0xE046),
        'segno': chr(0xE047),
        'coda': chr(0xE048),
        'codaSquare': chr(0xE049),
        'segnoSerpent2': chr(0xE04B),
        'gClef': chr(0xE050),
        'gClef15mb': chr(0xE051),
        'gClef8vb': chr(0xE052),
        'gClef8va': chr(0xE053),
        'gClef15ma': chr(0xE054),
        'gClef8vbOld': chr(0xE055),
        'gClef8vbCClef': chr(0xE056),
        'cClef': chr(0xE05C),
        'cClef8vb': chr(0xE05D),
        'fClef': chr(0xE062),
        'fClef15mb': chr(0xE063),
        'fClef8vb': chr(0xE064),
        'fClef8va': chr(0xE065),
        'fClef15ma': chr(0xE066),
        'unpitchedPercussionClef1': chr(0xE069),
        'unpitchedPercussionClef2': chr(0xE06A),
        '6stringTabClef': chr(0xE06D),
        '4stringTabClef': chr(0xE06E),
        'gClefReversed': chr(0xE073),
        'gClefTurned': chr(0xE074),
        'cClefReversed': chr(0xE075),
        'fClefReversed': chr(0xE076),
        'fClefTurned': chr(0xE077),
        'gClefChange': chr(0xE07A),
        'cClefChange': chr(0xE07B),
        'fClefChange': chr(0xE07C),
        'timeSig0': chr(0xE080),
        'timeSig1': chr(0xE081),
        'timeSig2': chr(0xE082),
        'timeSig3': chr(0xE083),
        'timeSig4': chr(0xE084),
        'timeSig5': chr(0xE085),
        'timeSig6': chr(0xE086),
        'timeSig7': chr(0xE087),
        'timeSig8': chr(0xE088),
        'timeSig9': chr(0xE089),
        'timeSigCommon': chr(0xE08A),
        'timeSigCutCommon': chr(0xE08B),
        'timeSigPlus': chr(0xE08C),
        'timeSigPlusSmall': chr(0xE08D),
        'timeSigFractionalSlash': chr(0xE08E),
        'timeSigEquals': chr(0xE08F),
        'timeSigMinus': chr(0xE090),
        'timeSigMultiply': chr(0xE091),
        'timeSigParensLeftSmall': chr(0xE092),
        'timeSigParensRightSmall': chr(0xE093),
        'timeSigParensLeft': chr(0xE094),
        'timeSigParensRight': chr(0xE095),
        'noteheadDoubleWhole': chr(0xE0A0),
        'noteheadDoubleWholeSquare': chr(0xE0A1),
        'noteheadWhole': chr(0xE0A2),
        'noteheadHalf': chr(0xE0A3),
        'noteheadBlack': chr(0xE0A4),
        'noteheadNull': chr(0xE0A5),
        'noteheadXBlack': chr(0xE0A9),
        'noteheadPlusBlack': chr(0xE0AF),
        'noteheadWholeWithX': chr(0xE0B5),
        'noteheadHalfWithX': chr(0xE0B6),
        'noteheadVoidWithX': chr(0xE0B7),
        'noteheadDiamondHalf': chr(0xE0D9),
        'noteheadDiamondHalfWide': chr(0xE0DA),
        'noteheadDiamondBlack': chr(0xE0DB),
        'noteheadDiamondBlackWide': chr(0xE0DC),
        'noteheadDiamondWhite': chr(0xE0DD),
        'noteheadDiamondWhiteWide': chr(0xE0DE),
        'noteheadParenthesisLeft': chr(0xE0F5),
        'noteheadParenthesisRight': chr(0xE0F6),
        'noteheadWholeFilled': chr(0xE0FA),
        'noteheadSlashHorizontalEnds': chr(0xE101),
        'noteheadSlashWhiteWhole': chr(0xE102),
        'noteheadSlashWhiteHalf': chr(0xE103),
        'noteDoubleWhole': chr(0xE1D0),
        'noteDoubleWholeSquare': chr(0xE1D1),
        'noteWhole': chr(0xE1D2),
        'noteHalfUp': chr(0xE1D3),
        'noteQuarterUp': chr(0xE1D5),
        'note8thUp': chr(0xE1D7),
        'note16thUp': chr(0xE1D9),
        'note32ndUp': chr(0xE1DB),
        'note64thUp': chr(0xE1DD),
        'note128thUp': chr(0xE1DF),
        'note256thUp': chr(0xE1E1),
        'note512thUp': chr(0xE1E3),
        'note1024thUp': chr(0xE1E5),
        'augmentationDot': chr(0xE1E7),
        'tremolo1': chr(0xE220),
        'tremolo2': chr(0xE221),
        'tremolo3': chr(0xE222),
        'tremolo4': chr(0xE223),
        'tremolo5': chr(0xE224),
        'buzzRoll': chr(0xE22A),
        'tremoloDivisiDots2': chr(0xE22E),
        'tremoloDivisiDots3': chr(0xE22F),
        'tremoloDivisiDots4': chr(0xE230),
        'tremoloDivisiDots6': chr(0xE231),
        'flag8thUp': chr(0xE240),
        'flag8thDown': chr(0xE241),
        'flag16thUp': chr(0xE242),
        'flag16thDown': chr(0xE243),
        'flag32ndUp': chr(0xE244),
        'flag32ndDown': chr(0xE245),
        'flag64thUp': chr(0xE246),
        'flag64thDown': chr(0xE247),
        'flag128thUp': chr(0xE248),
        'flag128thDown': chr(0xE249),
        'flag256thUp': chr(0xE24A),
        'flag256thDown': chr(0xE24B),
        'flag512thUp': chr(0xE24C),
        'flag512thDown': chr(0xE24D),
        'flag1024thUp': chr(0xE24E),
        'flag1024thDown': chr(0xE24F),
        'accidentalFlat': chr(0xE260),
        'accidentalNatural': chr(0xE261),
        'accidentalSharp': chr(0xE262),
        'accidentalDoubleSharp': chr(0xE263),
        'accidentalDoubleFlat': chr(0xE264),
        'accidentalTripleSharp': chr(0xE265),
        'accidentalTripleFlat': chr(0xE266),
        'accidentalNaturalFlat': chr(0xE267),
        'accidentalNaturalSharp': chr(0xE268),
        'accidentalSharpSharp': chr(0xE269),
        'accidentalParensLeft': chr(0xE26A),
        'accidentalParensRight': chr(0xE26B),
        'accidentalBracketLeft': chr(0xE26C),
        'accidentalBracketRight': chr(0xE26D),
        'accidentalQuarterToneFlatArrowUp': chr(0xE270),
        'accidentalThreeQuarterTonesFlatArrowDown': chr(0xE271),
        'accidentalQuarterToneSharpNaturalArrowUp': chr(0xE272),
        'accidentalQuarterToneFlatNaturalArrowDown': chr(0xE273),
        'accidentalThreeQuarterTonesSharpArrowUp': chr(0xE274),
        'accidentalQuarterToneSharpArrowDown': chr(0xE275),
        'accidentalQuarterToneFlatStein': chr(0xE280),
        'accidentalThreeQuarterTonesFlatZimmermann': chr(0xE281),
        'accidentalQuarterToneSharpStein': chr(0xE282),
        'accidentalThreeQuarterTonesSharpStein': chr(0xE283),
        'accidentalBuyukMucennebFlat': chr(0xE440),
        'accidentalKucukMucennebFlat': chr(0xE441),
        'accidentalBakiyeFlat': chr(0xE442),
        'accidentalKomaFlat': chr(0xE443),
        'accidentalKomaSharp': chr(0xE444),
        'accidentalBakiyeSharp': chr(0xE445),
        'accidentalKucukMucennebSharp': chr(0xE446),
        'accidentalBuyukMucennebSharp': chr(0xE447),
        'accidentalKoron': chr(0xE460),
        'accidentalSori': chr(0xE461),
        'articAccentAbove': chr(0xE4A0),
        'articAccentBelow': chr(0xE4A1),
        'articStaccatoAbove': chr(0xE4A2),
        'articStaccatoBelow': chr(0xE4A3),
        'articTenutoAbove': chr(0xE4A4),
        'articTenutoBelow': chr(0xE4A5),
        'articStaccatissimoAbove': chr(0xE4A6),
        'articStaccatissimoBelow': chr(0xE4A7),
        'articStaccatissimoWedgeAbove': chr(0xE4A8),
        'articStaccatissimoWedgeBelow': chr(0xE4A9),
        'articStaccatissimoStrokeAbove': chr(0xE4AA),
        'articStaccatissimoStrokeBelow': chr(0xE4AB),
        'articMarcatoAbove': chr(0xE4AC),
        'articMarcatoBelow': chr(0xE4AD),
        'articMarcatoStaccatoAbove': chr(0xE4AE),
        'articMarcatoStaccatoBelow': chr(0xE4AF),
        'articAccentStaccatoAbove': chr(0xE4B0),
        'articAccentStaccatoBelow': chr(0xE4B1),
        'articTenutoStaccatoAbove': chr(0xE4B2),
        'articTenutoStaccatoBelow': chr(0xE4B3),
        'articTenutoAccentAbove': chr(0xE4B4),
        'articTenutoAccentBelow': chr(0xE4B5),
        'articStressAbove': chr(0xE4B6),
        'articStressBelow': chr(0xE4B7),
        'articUnstressAbove': chr(0xE4B8),
        'articUnstressBelow': chr(0xE4B9),
        'articLaissezVibrerAbove': chr(0xE4BA),
        'articLaissezVibrerBelow': chr(0xE4BB),
        'articMarcatoTenutoAbove': chr(0xE4BC),
        'articMarcatoTenutoBelow': chr(0xE4BD),
        'fermataAbove': chr(0xE4C0),
        'fermataBelow': chr(0xE4C1),
        'fermataVeryShortAbove': chr(0xE4C2),
        'fermataVeryShortBelow': chr(0xE4C3),
        'fermataShortAbove': chr(0xE4C4),
        'fermataShortBelow': chr(0xE4C5),
        'fermataLongAbove': chr(0xE4C6),
        'fermataLongBelow': chr(0xE4C7),
        'fermataVeryLongAbove': chr(0xE4C8),
        'fermataVeryLongBelow': chr(0xE4C9),
        'breathMarkComma': chr(0xE4CE),
        'caesura': chr(0xE4D1),
        'caesuraThick': chr(0xE4D2),
        'caesuraShort': chr(0xE4D3),
        'caesuraCurved': chr(0xE4D4),
        'caesuraSingleStroke': chr(0xE4D7),
        'restMaxima': chr(0xE4E0),
        'restLonga': chr(0xE4E1),
        'restDoubleWhole': chr(0xE4E2),
        'restWhole': chr(0xE4E3),
        'restHalf': chr(0xE4E4),
        'restQuarter': chr(0xE4E5),
        'rest8th': chr(0xE4E6),
        'rest16th': chr(0xE4E7),
        'rest32nd': chr(0xE4E8),
        'rest64th': chr(0xE4E9),
        'rest128th': chr(0xE4EA),
        'rest256th': chr(0xE4EB),
        'rest512th': chr(0xE4EC),
        'rest1024th': chr(0xE4ED),
        'restQuarterOld': chr(0xE4F2),
        'restQuarterZ': chr(0xE4F6),
        'repeat1Bar': chr(0xE500),
        'repeat2Bars': chr(0xE501),
        'repeat4Bars': chr(0xE502),
        'repeatBarUpperDot': chr(0xE503),
        'repeatBarSlash': chr(0xE504),
        'repeatBarLowerDot': chr(0xE505),
        'ottava': chr(0xE510),
        'ottavaAlta': chr(0xE511),
        'ottavaBassa': chr(0xE512),
        'ottavaBassaBa': chr(0xE513),
        'quindicesima': chr(0xE514),
        'quindicesimaAlta': chr(0xE515),
        'quindicesimaBassa': chr(0xE516),
        'ventiduesima': chr(0xE517),
        'ventiduesimaAlta': chr(0xE518),
        'ventiduesimaBassa': chr(0xE519),
        'octaveParensLeft': chr(0xE51A),
        'octaveParensRight': chr(0xE51B),
        'ottavaBassaVb': chr(0xE51C),
        'quindicesimaBassaMb': chr(0xE51D),
        'ventiduesimaBassaMb': chr(0xE51E),
        'dynamicPiano': chr(0xE520),
        'dynamicMezzo': chr(0xE521),
        'dynamicForte': chr(0xE522),
        'dynamicRinforzando': chr(0xE523),
        'dynamicSforzando': chr(0xE524),
        'dynamicZ': chr(0xE525),
        'dynamicNiente': chr(0xE526),
        'dynamicPPPPPP': chr(0xE527),
        'dynamicPPPPP': chr(0xE528),
        'dynamicPPPP': chr(0xE529),
        'dynamicPPP': chr(0xE52A),
        'dynamicPP': chr(0xE52B),
        'dynamicMP': chr(0xE52C),
        'dynamicMF': chr(0xE52D),
        'dynamicPF': chr(0xE52E),
        'dynamicFF': chr(0xE52F),
        'dynamicFFF': chr(0xE530),
        'dynamicFFFF': chr(0xE531),
        'dynamicFFFFF': chr(0xE532),
        'dynamicFFFFFF': chr(0xE533),
        'dynamicFortePiano': chr(0xE534),
        'dynamicForzando': chr(0xE535),
        'dynamicSforzando1': chr(0xE536),
        'dynamicSforzandoPiano': chr(0xE537),
        'dynamicSforzandoPianissimo': chr(0xE538),
        'dynamicSforzato': chr(0xE539),
        'dynamicSforzatoPiano': chr(0xE53A),
        'dynamicSforzatoFF': chr(0xE53B),
        'dynamicRinforzando1': chr(0xE53C),
        'dynamicRinforzando2': chr(0xE53D),
        'lyricsElisionNarrow': chr(0xE550),
        'lyricsElision': chr(0xE551),
        'lyricsElisionWide': chr(0xE552),
        'ornamentTrill': chr(0xE566),
        'ornamentTurn': chr(0xE567),
        'ornamentTurnInverted': chr(0xE568),
        'ornamentTurnSlash': chr(0xE569),
        'ornamentTurnUp': chr(0xE56A),
        'ornamentTurnUpS': chr(0xE56B),
        'ornamentShortTrill': chr(0xE56C),
        'ornamentMordent': chr(0xE56D),
        'ornamentTremblement': chr(0xE56E),
        'ornamentHaydn': chr(0xE56F),
        'ornamentVerticalLine': chr(0xE583),
        'ornamentSchleifer': chr(0xE587),
        'ornamentZigZagLineNoRightEnd': chr(0xE59D),
        'ornamentZigZagLineWithRightEnd': chr(0xE59E),
        'ornamentPrecompSlide': chr(0xE5B0),
        'ornamentPrecompDescendingSlide': chr(0xE5B1),
        'ornamentPrecompAppoggTrill': chr(0xE5B2),
        'ornamentPrecompAppoggTrillSuffix': chr(0xE5B3),
        'ornamentPrecompTurnTrillDAnglebert': chr(0xE5B4),
        'ornamentPrecompSlideTrillDAnglebert': chr(0xE5B5),
        'ornamentPrecompSlideTrillMarpurg': chr(0xE5B6),
        'ornamentPrecompTurnTrillBach': chr(0xE5B7),
        'ornamentPrecompSlideTrillBach': chr(0xE5B8),
        'ornamentPrecompSlideTrillMuffat': chr(0xE5B9),
        'ornamentPrecompSlideTrillSuffixMuffat': chr(0xE5BA),
        'ornamentPrecompTrillSuffixDandrieu': chr(0xE5BB),
        'ornamentPrecompPortDeVoixMordent': chr(0xE5BC),
        'ornamentPrecompTrillWithMordent': chr(0xE5BD),
        'ornamentPrecompCadence': chr(0xE5BE),
        'ornamentPrecompCadenceWithTurn': chr(0xE5BF),
        'ornamentPrecompDoubleCadenceLowerPrefix': chr(0xE5C0),
        'ornamentPrecompCadenceUpperPrefix': chr(0xE5C1),
        'ornamentPrecompCadenceUpperPrefixTurn': chr(0xE5C2),
        'ornamentPrecompDoubleCadenceUpperPrefix': chr(0xE5C3),
        'ornamentPrecompDoubleCadenceUpperPrefixTurn': chr(0xE5C4),
        'ornamentPrecompMordentRelease': chr(0xE5C5),
        'ornamentPrecompMordentUpperPrefix': chr(0xE5C6),
        'ornamentPrecompInvertedMordentUpperPrefix': chr(0xE5C7),
        'ornamentPrecompTrillLowerSuffix': chr(0xE5C8),
        'brassMuteClosed': chr(0xE5E5),
        'brassMuteHalfClosed': chr(0xE5E6),
        'brassMuteOpen': chr(0xE5E7),
        'stringsDownBow': chr(0xE610),
        'stringsDownBowTurned': chr(0xE611),
        'stringsUpBow': chr(0xE612),
        'stringsUpBowTurned': chr(0xE613),
        'stringsHarmonic': chr(0xE614),
        'pluckedSnapPizzicatoBelow': chr(0xE630),
        'pluckedSnapPizzicatoAbove': chr(0xE631),
        'pluckedBuzzPizzicato': chr(0xE632),
        'pluckedLeftHandPizzicato': chr(0xE633),
        'pluckedDamp': chr(0xE638),
        'vocalSprechgesang': chr(0xE645),
        'keyboardPedalPed': chr(0xE650),
        'keyboardPedalP': chr(0xE651),
        'keyboardPedalUp': chr(0xE655),
        'keyboardPedalSost': chr(0xE659),
        'keyboardPedalS': chr(0xE65A),
        'pictBeaterSoftXylophoneUp': chr(0xE770),
        'pictBeaterSoftXylophoneDown': chr(0xE771),
        'pictBeaterSoftXylophoneRight': chr(0xE772),
        'pictBeaterSoftXylophoneLeft': chr(0xE773),
        'pictBeaterMediumXylophoneUp': chr(0xE774),
        'pictBeaterMediumXylophoneDown': chr(0xE775),
        'pictBeaterMediumXylophoneRight': chr(0xE776),
        'pictBeaterMediumXylophoneLeft': chr(0xE777),
        'pictBeaterHardXylophoneUp': chr(0xE778),
        'pictBeaterHardXylophoneDown': chr(0xE779),
        'pictBeaterHardXylophoneRight': chr(0xE77A),
        'pictBeaterHardXylophoneLeft': chr(0xE77B),
        'pictBeaterWoodXylophoneUp': chr(0xE77C),
        'pictBeaterWoodXylophoneDown': chr(0xE77D),
        'pictBeaterWoodXylophoneRight': chr(0xE77E),
        'pictBeaterWoodXylophoneLeft': chr(0xE77F),
        'pictBeaterSnareSticksUp': chr(0xE7D1),
        'pictBeaterSnareSticksDown': chr(0xE7D2),
        'pictBeaterHand': chr(0xE7E3),
        'pictBeaterFinger': chr(0xE7E4),
        'pictBeaterFist': chr(0xE7E5),
        'tuplet0': chr(0xE880),
        'tuplet1': chr(0xE881),
        'tuplet2': chr(0xE882),
        'tuplet3': chr(0xE883),
        'tuplet4': chr(0xE884),
        'tuplet5': chr(0xE885),
        'tuplet6': chr(0xE886),
        'tuplet7': chr(0xE887),
        'tuplet8': chr(0xE888),
        'tuplet9': chr(0xE889),
        'tupletColon': chr(0xE88A),
        'mensuralGclefPetrucci': chr(0xE901),
        'chantFclef': chr(0xE902),
        'mensuralFclef': chr(0xE903),
        'mensuralFclefPetrucci': chr(0xE904),
        'mensuralCclef': chr(0xE905),
        'chantCclef': chr(0xE906),
        'mensuralCclefPetrucciPosLowest': chr(0xE907),
        'mensuralCclefPetrucciPosLow': chr(0xE908),
        'mensuralCclefPetrucciPosMiddle': chr(0xE909),
        'mensuralCclefPetrucciPosHigh': chr(0xE90A),
        'mensuralCclefPetrucciPosHighest': chr(0xE90B),
        'mensuralProlation1': chr(0xE910),
        'mensuralProlation2': chr(0xE911),
        'mensuralProlation3': chr(0xE912),
        'mensuralProlation4': chr(0xE913),
        'mensuralProlation5': chr(0xE914),
        'mensuralProlation6': chr(0xE915),
        'mensuralProlation7': chr(0xE916),
        'mensuralProlation8': chr(0xE917),
        'mensuralProlation9': chr(0xE918),
        'mensuralProlation10': chr(0xE919),
        'mensuralProlation11': chr(0xE91A),
        'mensuralProportionTempusPerfectum': chr(0xE91B),
        'mensuralProportionProportioDupla1': chr(0xE91C),
        'mensuralProportionProportioDupla2': chr(0xE91D),
        'mensuralProportionProportioTripla': chr(0xE91E),
        'mensuralProportionProportioQuadrupla': chr(0xE91F),
        'mensuralProlationCombiningDot': chr(0xE920),
        'mensuralProlationCombiningTwoDots': chr(0xE921),
        'mensuralProlationCombiningThreeDots': chr(0xE922),
        'mensuralProlationCombiningThreeDotsTri': chr(0xE923),
        'mensuralProlationCombiningDotVoid': chr(0xE924),
        'mensuralProlationCombiningStroke': chr(0xE925),
        'mensuralNoteheadSemibrevisBlack': chr(0xE938),
        'mensuralNoteheadSemibrevisVoid': chr(0xE939),
        'mensuralNoteheadMinimaWhite': chr(0xE93C),
        'mensuralNoteheadSemiminimaWhite': chr(0xE93D),
        'mensuralCombStemUp': chr(0xE93E),
        'mensuralCombStemDown': chr(0xE93F),
        'mensuralCombStemDiagonal': chr(0xE940),
        'mensuralCombStemUpFlagRight': chr(0xE941),
        'mensuralCombStemDownFlagRight': chr(0xE942),
        'mensuralCombStemUpFlagLeft': chr(0xE943),
        'mensuralCombStemDownFlagLeft': chr(0xE944),
        'mensuralCombStemUpFlagFlared': chr(0xE945),
        'mensuralCombStemDownFlagFlared': chr(0xE946),
        'mensuralCombStemUpFlagExtended': chr(0xE947),
        'mensuralCombStemDownFlagExtended': chr(0xE948),
        'mensuralCombStemUpFlagSemiminima': chr(0xE949),
        'mensuralCombStemDownFlagSemiminima': chr(0xE94A),
        'mensuralCombStemUpFlagFusa': chr(0xE94B),
        'mensuralCombStemDownFlagFusa': chr(0xE94C),
        'chantPunctum': chr(0xE990),
        'chantPunctumInclinatum': chr(0xE991),
        'chantPunctumInclinatumAuctum': chr(0xE992),
        'chantPunctumInclinatumDeminutum': chr(0xE993),
        'chantAuctumAsc': chr(0xE994),
        'chantAuctumDesc': chr(0xE995),
        'chantPunctumVirga': chr(0xE996),
        'chantPunctumVirgaReversed': chr(0xE997),
        'chantQuilisma': chr(0xE99B),
        'chantOriscusLiquescens': chr(0xE99E),
        'chantStrophicus': chr(0xE99F),
        'chantPodatusLower': chr(0xE9B0),
        'chantPodatusUpper': chr(0xE9B1),
        'chantDeminutumUpper': chr(0xE9B2),
        'chantDeminutumLower': chr(0xE9B3),
        'chantEntryLineAsc2nd': chr(0xE9B4),
        'chantEntryLineAsc3rd': chr(0xE9B5),
        'chantEntryLineAsc4th': chr(0xE9B6),
        'chantEntryLineAsc5th': chr(0xE9B7),
        'chantEntryLineAsc6th': chr(0xE9B8),
        'chantLigaturaDesc2nd': chr(0xE9B9),
        'chantLigaturaDesc3rd': chr(0xE9BA),
        'chantLigaturaDesc4th': chr(0xE9BB),
        'chantLigaturaDesc5th': chr(0xE9BC),
        'chantConnectingLineAsc2nd': chr(0xE9BD),
        'chantConnectingLineAsc3rd': chr(0xE9BE),
        'chantConnectingLineAsc4th': chr(0xE9BF),
        'chantConnectingLineAsc5th': chr(0xE9C0),
        'chantConnectingLineAsc6th': chr(0xE9C1),
        'medRenFlatSoftB': chr(0xE9E0),
        'medRenFlatHardB': chr(0xE9E1),
        'medRenNatural': chr(0xE9E2),
        'medRenSharpCroix': chr(0xE9E3),
        'mensuralRestMaxima': chr(0xE9F0),
        'mensuralRestLongaPerfecta': chr(0xE9F1),
        'mensuralRestLongaImperfecta': chr(0xE9F2),
        'mensuralRestBrevis': chr(0xE9F3),
        'mensuralRestSemibrevis': chr(0xE9F4),
        'mensuralRestMinima': chr(0xE9F5),
        'mensuralRestSemiminima': chr(0xE9F6),
        'mensuralRestFusa': chr(0xE9F7),
        'mensuralRestSemifusa': chr(0xE9F8),
        'mensuralCustosUp': chr(0xEA02),
        'chantCustosStemUpPosMiddle': chr(0xEA06),
        'figbass0': chr(0xEA50),
        'figbass1': chr(0xEA51),
        'figbass2': chr(0xEA52),
        'figbass2Raised': chr(0xEA53),
        'figbass3': chr(0xEA54),
        'figbass4': chr(0xEA55),
        'figbass4Raised': chr(0xEA56),
        'figbass5': chr(0xEA57),
        'figbass5Raised1': chr(0xEA58),
        'figbass5Raised2': chr(0xEA59),
        'figbass5Raised3': chr(0xEA5A),
        'figbass6': chr(0xEA5B),
        'figbass6Raised': chr(0xEA5C),
        'figbass7': chr(0xEA5D),
        'figbass7Raised1': chr(0xEA5E),
        'figbass7Raised2': chr(0xEA5F),
        'figbass8': chr(0xEA60),
        'figbass9': chr(0xEA61),
        'figbass9Raised': chr(0xEA62),
        'figbassDoubleFlat': chr(0xEA63),
        'figbassFlat': chr(0xEA64),
        'figbassNatural': chr(0xEA65),
        'figbassSharp': chr(0xEA66),
        'figbassDoubleSharp': chr(0xEA67),
        'wiggleTrill': chr(0xEAA4),
        'wiggleArpeggiatoUp': chr(0xEAA9),
        'wiggleArpeggiatoDown': chr(0xEAAA),
        'wiggleArpeggiatoUpArrow': chr(0xEAAD),
        'wiggleArpeggiatoDownArrow': chr(0xEAAE),
        'wiggleGlissando': chr(0xEAAF),
        'elecPlay': chr(0xEB1C),
        'elecStop': chr(0xEB1D),
        'elecPause': chr(0xEB1E),
        'elecFastForward': chr(0xEB1F),
        'elecRewind': chr(0xEB20),
        'elecSkipForwards': chr(0xEB21),
        'elecSkipBackwards': chr(0xEB22),
        'luteDurationWhole': chr(0xEBA7),
        'luteDurationHalf': chr(0xEBA8),
        'luteDurationQuarter': chr(0xEBA9),
        'luteDuration8th': chr(0xEBAA),
        'luteDuration16th': chr(0xEBAB),
        'luteDuration32nd': chr(0xEBAC),
        'luteFrenchFretA': chr(0xEBC0),
        'luteFrenchFretB': chr(0xEBC1),
        'luteFrenchFretC': chr(0xEBC2),
        'luteFrenchFretD': chr(0xEBC3),
        'luteFrenchFretE': chr(0xEBC4),
        'luteFrenchFretF': chr(0xEBC5),
        'luteFrenchFretG': chr(0xEBC6),
        'luteFrenchFretH': chr(0xEBC7),
        'luteFrenchFretI': chr(0xEBC8),
        'luteFrenchFretK': chr(0xEBC9),
        'luteFrenchFretL': chr(0xEBCA),
        'luteFrenchFretM': chr(0xEBCB),
        'luteFrenchFretN': chr(0xEBCC),
        'luteItalianFret0': chr(0xEBE0),
        'luteItalianFret1': chr(0xEBE1),
        'luteItalianFret2': chr(0xEBE2),
        'luteItalianFret3': chr(0xEBE3),
        'luteItalianFret4': chr(0xEBE4),
        'luteItalianFret5': chr(0xEBE5),
        'luteItalianFret6': chr(0xEBE6),
        'luteItalianFret7': chr(0xEBE7),
        'luteItalianFret8': chr(0xEBE8),
        'luteItalianFret9': chr(0xEBE9),
        'timeSigBracketLeft': chr(0xEC80),
        'timeSigBracketRight': chr(0xEC81),
        'timeSigBracketLeftSmall': chr(0xEC82),
        'timeSigBracketRightSmall': chr(0xEC83),
        'timeSigSlash': chr(0xEC84),
        'timeSigCut2': chr(0xEC85),
        'timeSigCut3': chr(0xEC86),
        'figbass7Diminished': chr(0xECC0),
        'metNoteDoubleWhole': chr(0xECA0),
        'metNoteDoubleWholeSquare': chr(0xECA1),
        'metNoteWhole': chr(0xECA2),
        'metNoteHalfUp': chr(0xECA3),
        'metNoteQuarterUp': chr(0xECA5),
        'metNote8thUp': chr(0xECA7),
        'metNote16thUp': chr(0xECA9),
        'metNote32ndUp': chr(0xECAB),
        'metNote64thUp': chr(0xECAD),
        'metNote128thUp': chr(0xECAF),
        'metNote256thUp': chr(0xECB1),
        'metNote512thUp': chr(0xECB3),
        'metNote1024thUp': chr(0xECB5),
        'metAugmentationDot': chr(0xECB7),
        'articSoftAccentAbove': chr(0xED40),
        'articSoftAccentBelow': chr(0xED41),
    }

    SMUFL_METRONOME_MARK_NOTE_CHARS_TO_HUMDRUM_NOTE_NAME: dict[str, str] = {
        SMUFL_NAME_TO_UNICODE_CHAR['metNoteDoubleWhole']: 'double-whole',
        SMUFL_NAME_TO_UNICODE_CHAR['metNoteDoubleWholeSquare']: 'breve',
        SMUFL_NAME_TO_UNICODE_CHAR['metNoteWhole']: 'whole',
        SMUFL_NAME_TO_UNICODE_CHAR['metNoteHalfUp']: 'half',
        SMUFL_NAME_TO_UNICODE_CHAR['metNoteQuarterUp']: 'quarter',
        SMUFL_NAME_TO_UNICODE_CHAR['metNote8thUp']: '8th',
        SMUFL_NAME_TO_UNICODE_CHAR['metNote16thUp']: '16th',
        SMUFL_NAME_TO_UNICODE_CHAR['metNote32ndUp']: '32nd',
        SMUFL_NAME_TO_UNICODE_CHAR['metNote64thUp']: '64th',
        SMUFL_NAME_TO_UNICODE_CHAR['metNote128thUp']: '128th',
        SMUFL_NAME_TO_UNICODE_CHAR['metNote256thUp']: '256th',
        SMUFL_NAME_TO_UNICODE_CHAR['metNote512thUp']: '512th',
        SMUFL_NAME_TO_UNICODE_CHAR['metNote1024thUp']: '1024th'
    }

    SMUFL_METRONOME_MARK_NOTE_CHARS_TO_MUSIC21_FULL_NAME: dict[str, str] = {
        SMUFL_NAME_TO_UNICODE_CHAR['metNoteDoubleWhole']: 'Breve',
        SMUFL_NAME_TO_UNICODE_CHAR['metNoteWhole']: 'Whole',
        SMUFL_NAME_TO_UNICODE_CHAR['metNoteHalfUp']: 'Half',
        SMUFL_NAME_TO_UNICODE_CHAR['metNoteQuarterUp']: 'Quarter',
        SMUFL_NAME_TO_UNICODE_CHAR['metNote8thUp']: '8th',
        SMUFL_NAME_TO_UNICODE_CHAR['metNote16thUp']: '16th',
        SMUFL_NAME_TO_UNICODE_CHAR['metNote32ndUp']: '32nd',
        SMUFL_NAME_TO_UNICODE_CHAR['metNote64thUp']: '64th',
        SMUFL_NAME_TO_UNICODE_CHAR['metNote128thUp']: '128th',
        SMUFL_NAME_TO_UNICODE_CHAR['metNote256thUp']: '256th',
        SMUFL_NAME_TO_UNICODE_CHAR['metNote512thUp']: '512th',
        SMUFL_NAME_TO_UNICODE_CHAR['metNote1024thUp']: '1024th'
    }



# -----------------------------------------------------------------------------
# Name:         abc/__init__.py
# Purpose:      ABC converter internal interfaces
#
# Authors:      Greg Chapman
#
# Copyright:    Copyright © 2025 Greg Chapman
# License:      BSD, see license.txt
# -----------------------------------------------------------------------------

from .abcmetadata import AbcMetadata
from .abcreader import AbcReader
from .abcwriter import AbcWriter



# ------------------------------------------------------------------------------
# Name:          abcreader.py
# Purpose:       AbcReader reads an ABC file, and converts it to a music21 stream.
#
# Authors:       Greg Chapman <gregc@mac.com>
#
# Copyright:     (c) 2025 Greg Chapman
# License:       MIT, see LICENSE
# ------------------------------------------------------------------------------
import sys
import typing as t
import re

import music21 as m21

from converter21.shared import M21Utilities
from converter21.abc import abc2xml
from converter21.abc import AbcMetadata

class AbcImportException(Exception):
    pass


LIMIT_NEEDED: int = 10000

class AbcReader:
    def __init__(self, dataString: str):
        limit: int = sys.getrecursionlimit()
        if limit < LIMIT_NEEDED:
            sys.setrecursionlimit(LIMIT_NEEDED)
        self.abcString: str = dataString
        self.abcTuneByNumber: dict[str, str] = {}
        self.abcTuneHeaderLinesByNumber: dict[str, list[str]] = {}
        self.numberForAbcTune: dict[str, str] = {}
        self.abcTunesInDocumentOrder: list[str] = []

    def run(
        self,
        number: int | None
    ) -> m21.stream.Score | m21.stream.Part | m21.stream.Opus:
        # convert abc data to musicxml data and then import
        # _that_ into music21.

        xmlStrs: list[str] = []
        numStr: str
        if not self.abcTuneByNumber:
            # a bit of code stolen from abc2xml.getXmlDocs that finds the X:n number for
            # each tune (if there's more than one tune), and the associated tune.
            abctext: str = abc2xml.expand_abc_include(self.abcString)
            fragments: list[str] = re.split(r'^\s*X:', abctext, flags=re.M)
            preamble: str = fragments[0]
            tunes: list[str] = fragments[1:]
            if not tunes and preamble:
                tunes, preamble = ['1\n' + preamble], ''  # tune without X:

            self.abcTunesInDocumentOrder = []

            for tune in tunes:
                numberAndTuneRemainder: list[str] = tune.split('\n', 1)
                numStr = numberAndTuneRemainder[0].strip()
                fullTuneText: str = preamble + 'X:' + tune
                self.abcTunesInDocumentOrder.append(fullTuneText)
                self.abcTuneByNumber[numStr] = fullTuneText
                self.numberForAbcTune[fullTuneText] = numStr
                self.abcTuneHeaderLinesByNumber[numStr] = fullTuneText.split('\n')

        abcNumbers: list[str] = []
        if number is None:
            # all the tunes in the ABC data (in doc order)
            abcNumbers = []
            for tune in self.abcTunesInDocumentOrder:
                numStr = self.numberForAbcTune[tune]
                abcNumbers.append(numStr)
                xmlDoc = abc2xml.getXmlDocs(
                    self.abcTuneByNumber[numStr],
                    rOpt=True
                )[0]
                xmlStr: str = abc2xml.fixDoctype(xmlDoc)
                xmlStrs.append(xmlStr)
        else:
            numStr = str(number)
            if numStr not in self.abcTuneByNumber:
                raise AbcImportException(
                    f'cannot find requested reference number in source file: {number}'
                )
            abcNumbers = [numStr]
            xmlDoc = abc2xml.getXmlDocs(
                self.abcTuneByNumber[numStr],
                rOpt=True
            )[0]
            xmlStrs = [abc2xml.fixDoctype(xmlDoc)]

        if len(xmlStrs) == 1:
            # return a Score
            score = m21.converter.parseData(xmlStrs[0], fmt='musicxml')
            if t.TYPE_CHECKING:
                assert isinstance(score, m21.stream.Score)
            score.metadata = AbcMetadata.abcHeaderLinesToM21Metadata(
                self.abcTuneHeaderLinesByNumber[abcNumbers[0]]
            )
            mdNumbers = score.metadata['number']
            if len(mdNumbers) == 1 and str(mdNumbers[0]) == '1':
                # There's only one tune (len(xmlStrs) == 1), and it contained X:1,
                # which is meaningless, so delete that metadata item.
                score.metadata['number'] = None
            M21Utilities.fixupBadBeams(score, inPlace=True)
            return score

        # return an Opus of Scores, with each score.metadata.number set to the
        # abc tune reference number.
        opus = m21.stream.Opus()
        for xmlStr, numStr in zip(xmlStrs, abcNumbers):
            score = m21.converter.parseData(xmlStr)
            if t.TYPE_CHECKING:
                assert isinstance(score, m21.stream.Score)
            score.metadata = AbcMetadata.abcHeaderLinesToM21Metadata(
                self.abcTuneHeaderLinesByNumber[numStr]
            )
            opus.coreAppend(score)
        opus.coreElementsChanged()
        M21Utilities.fixupBadBeams(opus, inPlace=True)
        # if self.preambleHeaderLines:
        #     opus.metadata = AbcMetadata.abcHeaderLinesToM21Metadata(
        #         opus, self.preambleHeaderLines
        #     )
        return opus



# ------------------------------------------------------------------------------
# Name:          HumExceptions.py
# Purpose:       Exceptions that can be raised during Humdrum operations.
#
# Authors:       Greg Chapman <gregc@mac.com>
#                Humdrum code derived/translated from humlib (authored by
#                       Craig Stuart Sapp <craig@ccrma.stanford.edu>)
#
# Copyright:     (c) 2021-2023 Greg Chapman
# License:       MIT, see LICENSE
# ------------------------------------------------------------------------------
class HumdrumSyntaxError(Exception):
    # poorly formed humdrum input
    pass

class HumdrumInternalError(Exception):
    # unexpected state during import or export
    pass

class HumdrumExportError(Exception):
    # error converting something to humdrum
    pass



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Name:         mei/__init__.py
# Purpose:      MEI: The Music Encoding Initiative modules and classes
#
# Authors:      Greg Chapman
#
# Copyright:    Copyright © 2021-2025 Greg Chapman
# License:      BSD, see license.txt
# -----------------------------------------------------------------------------
'''
The :mod:`mei` module provides import from and export to MEI files for music21.
'''

from .meiexceptions import MeiValidityError
from .meiexceptions import MeiValueError
from .meiexceptions import MeiAttributeError
from .meiexceptions import MeiElementError
from .meiexceptions import MeiImportError
from .meiexceptions import MeiExportError
from .meiexceptions import MeiInternalError

from .m21objectconvert import M21ObjectConvert

from .meielement import MeiElement
from .meishared import MeiShared
from .meishared import MeiReaderScore
from .meimetadatareader import MeiMetadataReader
from .meireader import MeiReader
from .meireader import MEI_NS
from .meireader import INVALID_XML_DOC
from .meireader import WRONG_ROOT_ELEMENT

from .meimetadata import MeiMetadataItem
from .meimetadata import MeiMetadata
from .meilayer import MeiLayer
from .meistaff import MeiStaff
from .meimeasure import MeiMeasure
from .meiscore import MeiScore
from .meiwriter import MeiWriter



import pytest

# The things we're testing
import music21 as m21
from converter21.shared import M21Utilities

# test utilities
from tests.Utilities import *

def test_DateSingle():
    string = '///::.11'  # eleven one-hundredths of a second
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedSecond = 0)
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = '///::') # truncates the 11 milliseconds, so no seconds at all

    string = 'year/month/day/hour:minutes:.11'  # unparseable (but reasonably well-formed)
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    assert dateSingle is None # no elaborate check needed
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = '')

    string = '///::11' # 11th second
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedSecond = 11)
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = '///::11')

    string = '///:11' # 11th minute
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedMinute = 11)
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = '///:11:') # adds trailing ':'

    string = '///11' # 11 o’clock
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedHour = 11)
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = '///11::') # adds trailing ':'s

    string = '11' # the year 11 A.D.
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedYear = 11)
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = string)

    string = '11/' # the year 11 A.D.
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedYear = 11)
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = '11') # removes trailing '/'s

    string = '11//' # the year 11 A.D.
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedYear = 11)
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = '11') # removes trailing '/'s

    string = '11///' # the year 11 A.D.
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedYear = 11)
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = '11') # removes trailing '/'s

    string = '@11' # the year 11 B.C.E.
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedYear = -11)
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = string)

    string = '@11/' # the year 11 B.C.E.
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedYear = -11)
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = '@11') # removes trailing '/'

    string = '@11//' # the year 11 B.C.E.
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedYear = -11)
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = '@11') # removes trailing '/'s

    string = '@11///' # the year 11 B.C.E.
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedYear = -11)
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = '@11') # removes trailing '/'s

    string = '/11' # November
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedMonth = 11)
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = string)

    string = '//11' # 11th day of the month
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedDay = 11)
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = string)

    # more normal cases (just date, no time)
    string = '1943'
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedYear = 1943)
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = string)

    string = '1960/3/22'
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedYear = 1960,
                                   expectedMonth = 3,
                                   expectedDay = 22)
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = '1960/03/22') # adds leading 0 to month

    string = '@12345/10/31'
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedYear = -12345,
                                   expectedMonth = 10,
                                   expectedDay = 31)
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = string)

    # now with some approximations (~ for the whole date, x for a value within the date)
    # and uncertainty (? for the whole date, z for a value within the date)
    string = '~1979/06/01'
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedYear = 1979,
                                   expectedMonth = 6,
                                   expectedDay = 1,
                                   expectedHour = None,
                                   expectedRelevance = 'approximate')
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = string)

    string = '1979x/06z/01z'
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedYear = 1979,
                                   expectedMonth = 6,
                                   expectedDay = 1,
                                   expectedYearError = 'approximate',
                                   expectedMonthError = 'uncertain',
                                   expectedDayError = 'uncertain',
                                   expectedRelevance = 'certain')
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = string)

    string = '1979x/06z/01y' # that 'y' is not a valid error symbol
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    assert dateSingle is None
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = '')

    string = '~1979x/06z/01z'
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedYear = 1979,
                                   expectedMonth = 6,
                                   expectedDay = 1,
                                   expectedYearError = 'approximate',
                                   expectedMonthError = 'uncertain',
                                   expectedDayError = 'uncertain',
                                   expectedRelevance = 'approximate')
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = string)

    string = '?1979z/06x/01z'
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedYear = 1979,
                                   expectedMonth = 6,
                                   expectedDay = 1,
                                   expectedYearError = 'uncertain',
                                   expectedMonthError = 'approximate',
                                   expectedDayError = 'uncertain',
                                   expectedRelevance = 'uncertain')
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = string)

    string = '?1979z/06x/01z/17:03:58.999'
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedYear = 1979,
                                   expectedMonth = 6,
                                   expectedDay = 1,
                                   expectedHour = 17,
                                   expectedMinute = 3,
                                   expectedSecond = 58,  # truncated off the 999 milliseconds
                                   expectedYearError = 'uncertain',
                                   expectedMonthError = 'approximate',
                                   expectedDayError = 'uncertain',
                                   expectedRelevance = 'uncertain')
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = '?1979z/06x/01z/17:03:58')

    string = '?1979z/06x/01z/17:03:59.999'
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedYear = 1979,
                                   expectedMonth = 6,
                                   expectedDay = 1,
                                   expectedHour = 17,
                                   expectedMinute = 3,
                                   expectedSecond = 59,
                                   expectedYearError = 'uncertain',
                                   expectedMonthError = 'approximate',
                                   expectedDayError = 'uncertain',
                                   expectedRelevance = 'uncertain')
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = '?1979z/06x/01z/17:03:59')

    string = '1979z/06x/01z/17x:03z:59.999x'
    dateSingle = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSingle(dateSingle, expectedYear = 1979,
                                   expectedMonth = 6,
                                   expectedDay = 1,
                                   expectedHour = 17,
                                   expectedMinute = 3,
                                   expectedSecond = 59,
                                   expectedYearError = 'uncertain',
                                   expectedMonthError = 'approximate',
                                   expectedDayError = 'uncertain',
                                   expectedHourError = 'approximate',
                                   expectedMinuteError = 'uncertain',
                                   expectedSecondError = 'approximate',
                                   expectedRelevance = 'certain')
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = '1979z/06x/01z/17x:03z:59x')

    # test a DateSingle object that was created by hand to be problematic for the writer...
    # We have an invalid yearError, which we should ignore, since it has no associated
    # error symbol ('x' for 'approximate' or 'z' for 'uncertain', but nothing at all for
    # 'what-the-even-heck')
    date = m21.metadata.Date(year=1943, month=7, yearError='what-the-even-heck')
    dateSingle = m21.metadata.DateSingle(date)
    newString: str = M21Utilities.stringFromM21DateObject(dateSingle)
    CheckString(newString, expectedString = '1943/07')

def test_DateRelative():
    # We don't need lots of different date formats here (they're tested in test_DateSingle).
    # Just a few examples of '<' and '>' dates.
    string = '>1979z/06x/01z/17x:03z:59.999x'
    dateRelative = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateRelative(dateRelative, expectedYear = 1979,
                                   expectedMonth = 6,
                                   expectedDay = 1,
                                   expectedHour = 17,
                                   expectedMinute = 3,
                                   expectedSecond = 59,
                                   expectedYearError = 'uncertain',
                                   expectedMonthError = 'approximate',
                                   expectedDayError = 'uncertain',
                                   expectedHourError = 'approximate',
                                   expectedMinuteError = 'uncertain',
                                   expectedSecondError = 'approximate',
                                   expectedRelevance = 'after')
    newString: str = M21Utilities.stringFromM21DateObject(dateRelative)
    CheckString(newString, expectedString = '>1979z/06x/01z/17x:03z:59x')

    string = '<1979z/06x/01z/17x:03z:59.999x'
    dateRelative = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateRelative(dateRelative, expectedYear = 1979,
                                   expectedMonth = 6,
                                   expectedDay = 1,
                                   expectedHour = 17,
                                   expectedMinute = 3,
                                   expectedSecond = 59,
                                   expectedYearError = 'uncertain',
                                   expectedMonthError = 'approximate',
                                   expectedDayError = 'uncertain',
                                   expectedHourError = 'approximate',
                                   expectedMinuteError = 'uncertain',
                                   expectedSecondError = 'approximate',
                                   expectedRelevance = 'prior')
    newString: str = M21Utilities.stringFromM21DateObject(dateRelative)
    CheckString(newString, expectedString = '<1979z/06x/01z/17x:03z:59x')

    # Try one with a '<' in the wrong location to make sure it doesn't parse
    string = '1979</06/01/17:03:56'
    dateRelative = M21Utilities.m21DatePrimitiveFromString(string)
    assert dateRelative is None

    # Try one with a '>' in the wrong location to make sure it doesn't parse
    string = '1979/06/01/17:03:56>'
    dateRelative = M21Utilities.m21DatePrimitiveFromString(string)
    assert dateRelative is None

def test_DateBetween():
    # a few straightforward examples (using both '-' and '^'), then some malformed ones (e.g.
    # 3 dates instead of two, or two dates, one of which is malformed, or 1 date)
    string = '1979/06/01/17:03:56-1979/06/30/17:03:56' # June 1-30, 1979 (timestamps equal)
    dateBetween = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateBetween(dateBetween, expectedYear = 1979,
                                   expectedMonth = 6,
                                   expectedDay = (1, 30),
                                   expectedHour = 17,
                                   expectedMinute = 3,
                                   expectedSecond = 56)
    newString: str = M21Utilities.stringFromM21DateObject(dateBetween)
    CheckString(newString, expectedString = string)

    string = '1979/06/01/17:03:56^1979/06/30/17:03:56' # June 1-30, 1979 (timestamps equal)
    dateBetween = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateBetween(dateBetween, expectedYear = 1979,
                                   expectedMonth = 6,
                                   expectedDay = (1, 30),
                                   expectedHour = 17,
                                   expectedMinute = 3,
                                   expectedSecond = 56)
    newString: str = M21Utilities.stringFromM21DateObject(dateBetween)
    CheckString(newString, expectedString = '1979/06/01/17:03:56-1979/06/30/17:03:56') # ^ -> -

    # this one has a backward range.  Apparently no-one cares...
    string = '1979/06/30/17:03:56-1979/06/01/17:03:56' # June 30-1, 1979 (timestamps equal)
    dateBetween = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateBetween(dateBetween, expectedYear = 1979,
                                   expectedMonth = 6,
                                   expectedDay = (30, 1),
                                   expectedHour = 17,
                                   expectedMinute = 3,
                                   expectedSecond = 56)
    newString: str = M21Utilities.stringFromM21DateObject(dateBetween)
    CheckString(newString, expectedString = string)

    # very simple example
    string = '1942-1943'
    dateBetween = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateBetween(dateBetween, expectedYear = (1942, 1943))
    newString: str = M21Utilities.stringFromM21DateObject(dateBetween)
    CheckString(newString, expectedString = string)

    # very simple bad example (3 dates)
    string = '1942-1943-1944'
    dateBetween = M21Utilities.m21DatePrimitiveFromString(string)
    assert dateBetween is None # should fail because '1943-1944' can't parse as a year number

    # very simple bad example (1 date)
    string = '1942-'
    dateBetween = M21Utilities.m21DatePrimitiveFromString(string)
    assert dateBetween is None # should fail because '' has no numeric date fields in it

def test_DateSelection():
    # a few straightforward examples, then some malformed ones
    string = '1979/06/01/17:03:56|1979/06/30/17:03:56' # June 1 or June 30, 1979 (timestamps equal)
    dateSelection = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSelection(dateSelection, expectedNumDates = 2,
                                   expectedYear = 1979,
                                   expectedMonth = 6,
                                   expectedDay = (1, 30),
                                   expectedHour = 17,
                                   expectedMinute = 3,
                                   expectedSecond = 56)
    newString: str = M21Utilities.stringFromM21DateObject(dateSelection)
    CheckString(newString, expectedString = string)

    # June 1, June 15, or June 30, 1979 (timestamps equal)
    string = '1979/06/01/17:03:56|1979/06/15/17:03:56|1979/06/30/17:03:56'
    dateSelection = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSelection(dateSelection, expectedNumDates = 3,
                                   expectedYear = 1979,
                                   expectedMonth = 6,
                                   expectedDay = (1, 15, 30),
                                   expectedHour = 17,
                                   expectedMinute = 3,
                                   expectedSecond = 56)
    newString: str = M21Utilities.stringFromM21DateObject(dateSelection)
    CheckString(newString, expectedString = string)

    string = '1940 | @100 | 1765 | @10000 | 1960 | 1961|1978|1979| 1995'
    dateSelection = M21Utilities.m21DatePrimitiveFromString(string)
    CheckM21DateSelection(dateSelection, expectedNumDates = 9,
        expectedYear = (1940, -100, 1765, -10000, 1960, 1961, 1978, 1979, 1995) )
    newString: str = M21Utilities.stringFromM21DateObject(dateSelection)
    CheckString(newString, expectedString = '1940|@100|1765|@10000|1960|1961|1978|1979|1995') # no spaces




# ------------------------------------------------------------------------------
# Name:          HumHash.py
# Purpose:       A dictionary (with three levels of keys: namespace1, namespace2, key).
#                Each HumdrumToken, HumdrumLine, and HumdrumFile is derived from (has a)
#                HumHash, to hold the interesting info we have analyzed about that object.
#
# Authors:       Greg Chapman <gregc@mac.com>
#                Humdrum code derived/translated from humlib (authored by
#                       Craig Stuart Sapp <craig@ccrma.stanford.edu>)
#
# Copyright:     (c) 2021-2023 Greg Chapman
# License:       MIT, see LICENSE
# ------------------------------------------------------------------------------

import typing as t

from music21 import Music21Object
from music21.common import opFrac
from music21.common import OffsetQL

from converter21.humdrum import HumdrumInternalError

def getKeyTuple(inKey: str) -> tuple[str, str, str]:
    keyList: list[str] = inKey.split(':')
    if len(keyList) == 1:
        return ('', '', keyList[0])

    if len(keyList) == 2:
        return ('', keyList[0], keyList[1])

    return (keyList[0], keyList[1], keyList[2])

def fixupNamespace1Namespace2Key(*ns1ns2key: str) -> tuple[str, str, str]:
    if len(ns1ns2key) < 1:
        raise ValueError("too few specifiers (need at least a key)")
    if len(ns1ns2key) > 3:
        raise ValueError("too many specifiers (should be at most 3: ns1, ns2, key)")

    ns1 = ''
    ns2 = ''

    if len(ns1ns2key) == 1:
        key = ns1ns2key[0]
    elif len(ns1ns2key) == 2:
        ns2 = ns1ns2key[0]
        key = ns1ns2key[1]
    else:  # len(ns1ns2key) == 3
        ns1 = ns1ns2key[0]
        ns2 = ns1ns2key[1]
        key = ns1ns2key[2]

    # If ns1 and ns2 are not specified, key might have colon-delimited
    # namespaces in it, so we need to parse them out.
    if ns1 == '' and ns2 == '':
        ns1, ns2, key = getKeyTuple(key)

    # We support single level namespace key/value pairs that have either
    # ns1 == '' or ns2 == '' as equivalent. In other words, they both will
    # land in {'', {singleNS, {key, value}}}.
    if ns2 == '' and ns1 != '':
        # swap them, for equivalency (make ns1 the empty string instead of ns2)
        ns1, ns2 = ns2, ns1

    return (ns1, ns2, key)

def fixupNamespace1Namespace2(
        ns1: str | None = None,
        ns2: str | None = None
) -> tuple[str | None, str | None]:
    if ns1 is None and ns2 is not None:
        ns1, ns2 = ns2, ns1  # make ns2 be the one that is None, if one of them is

    if ns1 is not None and ns2 is None:
        # check for a colon in ns1 -> generate ns1, ns2 from ns2 in that case
        keyTuple = getKeyTuple(ns1)
        if len(keyTuple) == 2:
            ns1 = keyTuple[0]
            ns2 = keyTuple[1]

    return (ns1, ns2)

class HumParameter:
    def __init__(self, value: t.Any | None, origin=None) -> None:
        # value can be of any type (different from humlib, where it's always a string)
        from converter21.humdrum import HumdrumToken
        if origin is not None and not isinstance(origin, HumdrumToken):
            raise HumdrumInternalError('HumParameter origin must be HumdrumToken | None')
        self._value: t.Any | None = value
        self._origin: HumdrumToken | None = origin

    @property
    def value(self) -> t.Any | None:
        return self._value

    @value.setter
    def value(self, newValue: t.Any | None) -> None:
        self._value = newValue

    @property
    def origin(self):  # -> HumdrumToken | None:
        return self._origin

    @origin.setter
    def origin(self, newOrigin) -> None:  # newOrigin: HumdrumToken | None
        from converter21.humdrum import HumdrumToken
        if newOrigin is not None and not isinstance(newOrigin, HumdrumToken):
            raise HumdrumInternalError('invalid newOrigin')
        self._origin = newOrigin

class HumHash:
    def __init__(self) -> None:
        # {ns1...,{ns2..., {key..., value...}}}
        self._parameters: dict[str, dict[str, dict[str, HumParameter]]] = {}
        self._prefix: str = ''

    '''
    //////////////////////////////
    //
    // HumHash::setPrefix -- initial string to print when using
    //   operator<<.  This is used for including the "!" for local
    //   comments or "!!" for global comments.   The prefix will
    //   remain the same until it is changed.  The default prefix
    //   of the object it the empty string.
    '''
    @property
    def prefix(self) -> str:
        return self._prefix

    @prefix.setter
    def prefix(self, newPrefix: str) -> None:
        self._prefix = newPrefix

    '''
    //////////////////////////////
    //
    // HumHash::setValue -- Set the parameter to the given value,
    //     over-writing any previous value for the parameter.  The
    //     value is any arbitrary string, but preferably does not
    //     include tabs or colons.  If a colon is needed, then specify
    //     as "&colon;" without the quotes.  Values such as integers
    //     fractions and floats can be specified, and these wil be converted
    //     internally into strings (use getValueInt() or getValueFloat()
    //     to recover the original type).
        GregC says: I have redefined this to store the value in its original
        form, and getValueString() if necessary (which might fail and return
        None).  Python (and I) cannot tolerate the HumdrumToken <->
        "HT_<64-bit-pointer>" string conversions, so I disallow that, and
        I hope that storing a reference to the actual HumdrumToken instead
        is a much better choice.
        Q: Perhaps we should store a weak reference to the HumdrumToken value
        Q: instead, to avoid circular references?
    '''
    def setValue(self, *ns1ns2keyvalue) -> None:
        value: t.Any | None = ns1ns2keyvalue[-1]       # value is last
        ns1ns2key = ns1ns2keyvalue[:-1]  # ns1ns2key is all but last
        ns1: str
        ns2: str
        key: str
        ns1, ns2, key = fixupNamespace1Namespace2Key(*ns1ns2key)

        if ns1 not in self._parameters:
            self._parameters[ns1] = dict([(ns2, dict([(key, HumParameter(value))]))])
        elif ns2 not in self._parameters[ns1]:
            self._parameters[ns1][ns2] = dict([(key, HumParameter(value))])
        else:
            self._parameters[ns1][ns2][key] = HumParameter(value)

    '''
    //////////////////////////////
    //
    // HumHash::getValue -- Returns the value specified by the given key.
    //    If there is no colon in the key then return the value for the key
    //    in the default namespaces (NS1="" and NS2="").  If there is one colon,
    //    then the two pieces of the string as NS2 and the key, with NS1="".
    //    If there are two colons, then that specified the complete namespaces/key
    //    address of the value.  The namespaces and key can be specified as
    //    separate parameters in a similar manner to the single-string version.
    //    But in these cases colon concatenation of the namespaces and/or key
    //    are not allowed.
    '''
    def getValue(self, *ns1ns2key) -> t.Any | None:
        if not self._parameters:
            return None

        ns1, ns2, key = fixupNamespace1Namespace2Key(*ns1ns2key)
        if ns1 not in self._parameters:
            return None
        if ns2 not in self._parameters[ns1]:
            return None
        if key not in self._parameters[ns1][ns2]:
            return None

        return self._parameters[ns1][ns2][key].value

    # getValue variants that convert to requested type (might return None if conversion fails)

    def getValueString(self, *ns1ns2key) -> str | None:
        value = self.getValue(*ns1ns2key)
        if value is None:
            return None

        from converter21.humdrum import HumdrumToken
        if isinstance(value, HumdrumToken):
            return None

        if isinstance(value, Music21Object):
            return None

        try:
            return str(value)  # can convert from int, HumNum, Fraction, float, str, etc
        except Exception:
            return None

    def getValueToken(self, *ns1ns2key):  # -> HumdrumToken | None:
        value = self.getValue(*ns1ns2key)
        if value is None:
            return None

        from converter21.humdrum import HumdrumToken
        if isinstance(value, HumdrumToken):
            return value

        return None  # can't convert to anything else

    def getValueM21Object(self, *ns1ns2key) -> Music21Object | None:
        value = self.getValue(*ns1ns2key)
        if value is None:
            return None

        if isinstance(value, Music21Object):
            return value

        return None  # can't convert to anything else

    def getValueInt(self, *ns1ns2key) -> int:
        value = self.getValue(*ns1ns2key)
        if value is None:
            return 0

        from converter21.humdrum import HumdrumToken
        if isinstance(value, HumdrumToken):
            return 0

        try:
            return int(value)  # can convert from int, float, HumNum, Fraction, str
        except Exception:
            return 0

    def getValueHumNum(self, *ns1ns2key) -> OffsetQL:  # -> HumNum:
        value = self.getValue(*ns1ns2key)
        if value is None:
            return opFrac(0)

        from converter21.humdrum import HumdrumToken
        if isinstance(value, HumdrumToken):
            return opFrac(0)

        try:
            return opFrac(value)  # can convert from int, float, Fraction, str
        except Exception:
            return opFrac(0)

    def getValueFloat(self, *ns1ns2key) -> float:
        value = self.getValue(*ns1ns2key)
        if value is None:
            return 0.0

        from converter21.humdrum import HumdrumToken
        if isinstance(value, HumdrumToken):
            return 0.0

        try:
            return float(value)  # can convert from int, HumNum, Fraction, str
        except Exception:
            return 0.0

    def getValueBool(self, *ns1ns2key) -> bool:
        # this one's weird.  We default to True (unless there's some fairly
        # obvious reason to interpret as False, e.g. string == '0', int = 0, etc)
        value = self.getValue(*ns1ns2key)
        if value is None:
            return False

        if isinstance(value, bool):
            return value

        from converter21.humdrum import HumdrumToken
        if isinstance(value, HumdrumToken):
            return True  # it's not None, so...

        if isinstance(value, str):
            if value == 'false':
                return False
            if value == '0':
                return False
            return True

        try:
            return int(value) != 0
        except Exception:
            return True  # it's not None, so...

    '''
    //////////////////////////////
    //
    // HumHash::isDefined -- Returns true if the given parameter exists in the
    //    map.   Format of the input string:   NS1:NS2:key or "":NS2:key for the
    //    two argument version of the function.  OR "":"":key if no colons in
    //    single string argument version.
    '''
    def isDefined(self, *ns1ns2key) -> bool:
        if not self._parameters:
            return False

        ns1, ns2, key = fixupNamespace1Namespace2Key(*ns1ns2key)
        if ns1 not in self._parameters:
            return False
        if ns2 not in self._parameters[ns1]:
            return False

        return key in self._parameters[ns1][ns2]

    '''
    //////////////////////////////
    //
    // HumHash::deleteValue -- Delete the given parameter key from the HumHash
    //   object.  Three string version is N1,NS2,key; two string version is
    //   "",NS2,key; and one argument version is "","",key.
    '''
    def deleteValue(self, *ns1ns2key) -> None:
        if not self._parameters:
            return

        ns1, ns2, key = fixupNamespace1Namespace2Key(*ns1ns2key)

        if ns1 not in self._parameters:
            return

        if ns2 not in self._parameters[ns1]:
            return

        if key in self._parameters[ns1][ns2]:
            del self._parameters[ns1][ns2][key]

    '''
    //////////////////////////////
    //
    // HumHash::setOrigin -- Set the source token for the parameter.
    '''
    def setOrigin(self, *ns1ns2keyvalue):  # origin: HumdrumToken
        if not self._parameters:
            return

        origin = ns1ns2keyvalue[-1]
        ns1ns2key = ns1ns2keyvalue[:-1]
        ns1, ns2, key = fixupNamespace1Namespace2Key(*ns1ns2key)

        from converter21.humdrum import HumdrumToken
        if not isinstance(origin, HumdrumToken):
            raise HumdrumInternalError('invalid origin token in HumHash.setOrigin')

        if ns1 not in self._parameters:
            return

        if ns2 not in self._parameters[ns1]:
            return

        if key not in self._parameters[ns1][ns2]:
            return

        self._parameters[ns1][ns2][key].origin = origin

    '''
    //////////////////////////////
    //
    // HumHash::getOrigin -- Get the source token for the parameter.
    //    Returns NULL if there is no origin.
    '''
    def getOrigin(self, *ns1ns2key):  # -> HumdrumToken:
        if not self._parameters:
            return None

        ns1, ns2, key = fixupNamespace1Namespace2Key(*ns1ns2key)

        if ns1 not in self._parameters:
            return None

        if ns2 not in self._parameters[ns1]:
            return None

        if key not in self._parameters[ns1][ns2]:
            return None

        return self._parameters[ns1][ns2][key].origin

    '''
    //////////////////////////////
    //
    // HumHash::getKeys -- Return a list of keys in a particular namespace
    //     combination.  With no parameters, a complete list of all
    //     namespaces/keys will be returned.  Giving one parameter will
    //     produce a list will give all NS2:key values in the NS1 namespace.
    //     If there is a colon in the single parameter version of the function,
    //     then this will be interpreted as "NS1", "NS2" version of the parameters
    //     described above.
    '''
    def getKeys(self, ns1: str | None = None, ns2: str | None = None) -> list[str]:
        retKeys: list[str] = []
        if not self._parameters:
            # return empty list of keys
            return retKeys

        ns1, ns2 = fixupNamespace1Namespace2(ns1, ns2)

        if ns1 is None and ns2 is None:
            # return all 'namespace1:namespace2:key's in self._parameters
            for my1 in self._parameters:
                for my2 in self._parameters[my1]:
                    for key in self._parameters[my1][my2]:
                        retKeys.append(my1 + ':' + my2 + ':' + key)
            return retKeys

        if ns2 is None:
            # return all 'namespace2:key's in the specified ns1 namespace
            if ns1 in self._parameters:
                for my2 in self._parameters[ns1]:
                    for key in self._parameters[ns1][my2]:
                        retKeys.append(my2 + ':' + key)
            return retKeys

        # return all keys in the specified ns1,ns2 namespace
        if ns1 in self._parameters:
            if ns2 in self._parameters[ns1]:
                for key in self._parameters[ns1][ns2]:
                    retKeys.append(key)
        return retKeys

    '''
    //////////////////////////////
    //
    // HumHash::hasParameters -- Returns true if at least one parameter is defined
    //     in the HumHash object (when no arguments are given to the function).
    //     When two strings are given as arguments, the function checks to see if
    //     the given namespace pair has any keys.  If only one string argument,
    //     then check if the given NS1 has any parameters, unless there is a
    //     colon in the string which means to check NS1:NS2.
    '''
    def hasParameters(self, ns1: str | None = None, ns2: str | None = None) -> bool:
        if not self._parameters:
            # return empty list of keys
            return False

        ns1, ns2 = fixupNamespace1Namespace2(ns1, ns2)

        if ns1 is None and ns2 is None:
            # return whether there are any parameters in self._parameters
            for my1 in self._parameters:
                for my2 in self._parameters[my1]:
                    if self._parameters[my1][my2]:
                        return True
            return False

        if ns2 is None:
            # return whether there are any parameters in the specified ns1 namespace
            if ns1 in self._parameters:
                for my2 in self._parameters[ns1]:
                    if self._parameters[ns1][my2]:
                        return True
            return False

        # return whether there are any parameters in the specified ns1,ns2 namespace
        if ns1 in self._parameters:
            if ns2 in self._parameters[ns1]:
                if self._parameters[ns1][ns2]:
                    return True
        return False

    '''
    //////////////////////////////
    //
    // HumHash::getParameterCount -- Return a count of the parameters which are
    //     stored in the HumHash.  If no arguments, then count all value in
    //     all namespaces.  If two arguments, then return the count for a
    //     specific NS1:NS2 namespace.  If one argument, then return the
    //     parameters in NS1, but if there is a colon in the string,
    //     return the parameters in NS1:NS2.
    //
    '''
    def getParameterCount(self, ns1: str | None = None, ns2: str | None = None) -> int:
        if not self._parameters:
            return 0

        ns1, ns2 = fixupNamespace1Namespace2(ns1, ns2)

        count = 0
        if ns1 is None and ns2 is None:
            # return count of parameters in self._parameters
            for my1 in self._parameters:
                for my2 in self._parameters[my1]:
                    count += len(self._parameters[my1][my2])
            return count

        count = 0
        if ns2 is None:
            # return count of parameters in the specified ns1 namespace
            if ns1 in self._parameters:
                for my2 in self._parameters[ns1]:
                    count += len(self._parameters[ns1][my2])
            return count

        # return count of parameters in the specified ns1,ns2 namespace
        if ns1 in self._parameters:
            if ns2 in self._parameters[ns1]:
                count = len(self._parameters[ns1][ns2])
        return count



# ------------------------------------------------------------------------------
# Name:          meiscore.py
# Purpose:       MeiScore is an object that takes a music21 Score and
#                organizes its objects in an MEI-like structure.
#
# Authors:       Greg Chapman <gregc@mac.com>
#
# Copyright:     (c) 2023-2025 Greg Chapman
# License:       MIT, see LICENSE
# ------------------------------------------------------------------------------
import sys
import typing as t
from xml.etree.ElementTree import Element  # , TreeBuilder

import music21 as m21
from music21.common import opFrac
from music21.common import OffsetQL

# from converter21.mei import MeiExportError
from converter21.mei import MeiInternalError
from converter21.mei import MeiMetadataItem
from converter21.mei import MeiMetadata
from converter21.mei import MeiMeasure
from converter21.mei import M21ObjectConvert
from converter21.shared import M21TemporarySpanner
from converter21.shared import M21BeamSpanner
from converter21.shared import M21TupletSpanner
from converter21.shared import M21TieSpanner
from converter21.shared import M21Utilities
from converter21.shared import M21StaffGroupTree
from converter21.shared import DebugTreeBuilder as TreeBuilder  # put this back before shipping

environLocal = m21.environment.Environment('converter21.mei.meiscore')

# For debug or unit test print, a simple way to get a string which is the current function name
# with a colon appended.
# for current func name, specify 0 or no argument.
# for name of caller of current func, specify 1.
# for name of caller of caller of current func, specify 2. etc.
# pylint: disable=protected-access
funcName = lambda n=0: sys._getframe(n + 1).f_code.co_name + ':'  # pragma no cover
# pylint: enable=protected-access

class MeiScore:
    def __init__(self, m21Score: m21.stream.Score, meiVersion: str) -> None:
        def getUniqueVoiceIds(part: m21.stream.Part) -> list[int | str]:
            output: list[int | str] = ['']  # fake id for Measure
            for meas in part[m21.stream.Measure]:
                for voice in meas.voices:
                    if voice.id not in output:
                        output.append(voice.id)
            return output

        self.m21Score: m21.stream.Score = m21Score
        self.meiVersion: str = meiVersion

        self.customM21AttrsToDelete: dict[m21.base.Music21Object, list[str]] = {}

        self.currentTupletSpanners: dict[m21.stream.Part, list[M21TupletSpanner]] = {}
        self.currentTieSpanners: dict[m21.stream.Part, list[tuple[M21TieSpanner, int]]] = {}

        # we assume beams do not cross voice ids within the part. (This will still be true
        # when we eventually support cross-part beaming... the voice id will still match.)
        self.previousBeamedNoteOrChord: dict[
            tuple[m21.stream.Part, int | str],
            m21.note.NotRest | None
        ] = {}
        self.currentBeamSpanners: dict[
            tuple[m21.stream.Part, int | str],
            list[M21BeamSpanner]
        ] = {}

        for part in self.m21Score.parts:
            self.currentTupletSpanners[part] = []
            self.currentTieSpanners[part] = []
            voiceIds: list[int | str] = getUniqueVoiceIds(part)
            for voiceId in voiceIds:
                self.currentBeamSpanners[(part, voiceId)] = []
                self.previousBeamedNoteOrChord[(part, voiceId)] = None

        # pre-scan of m21Score to set up some things
        self.annotateScore()

        # annotateScore put a bunch of stuff in the spannerBundle, so we can't
        # copy it until now.  Which is just in time, since makeXmlIds/assureAllXmlIds needs it.
        self.spannerBundle = self.m21Score.spannerBundle

        # once annotated, check to see if any of these objects need xml:id.
        # self.makeXmlIds()

        # Actually, put an xml:id on every single Music21Object in the score.
        M21Utilities.assureAllXmlIds(self.m21Score)

        self.scoreMeterStream: m21.stream.Stream[m21.meter.TimeSignature] = (
            self.m21Score.getTimeSignatures(
                returnDefault=True,
                searchContext=False,
                sortByCreationTime=False
            )
        )

        self.staffNumbersForM21Parts: dict[m21.stream.Part, int] = (
            self._getStaffNumbersForM21Parts()
        )
        self.m21PartsForStaffNumbers: dict[int, m21.stream.Part] = {
            value: key for key, value in self.staffNumbersForM21Parts.items()
        }

        self.staffGroupTrees: list[M21StaffGroupTree] = (
            M21Utilities.getStaffGroupTrees(
                list(m21Score[m21.layout.StaffGroup]),
                self.staffNumbersForM21Parts,
            )
        )

        self.measures: list[MeiMeasure] = self._getMeiMeasures()
        self.metadata: MeiMetadata = MeiMetadata(m21Score.metadata)

    def _getStaffNumbersForM21Parts(self) -> dict[m21.stream.Part, int]:
        # If this routine ever changes the fact that staff numbers are in the same
        # monotonically increasing order as m21Score.parts, starting at 1, incrementing
        # by 1 for each Part, then makeStaffGrpOrStaffDefElementStartingAt() will need
        # to be updated to match (it counts on that fact).
        output: dict[m21.stream.Part, int] = {}
        for staffIdx, part in enumerate(self.m21Score.parts):
            output[part] = staffIdx + 1  # staff numbers are 1-based
        return output

    def _getMeiMeasures(self) -> list[MeiMeasure]:
        output: list[MeiMeasure] = []

        measureStacks: list[list[m21.stream.Measure]] = []
        parts: list[m21.stream.Part] = list(self.m21Score.parts)
        partMeasures: list[list[m21.stream.Measure]] = []
        for part in parts:
            partMeasures.append(list(part.getElementsByClass(m21.stream.Measure)))

        # invert partMeasures into measureStacks
        measureStacks = [list(measureStack) for measureStack in zip(*partMeasures)]

        meiMeas: MeiMeasure | None = None
        for measureStack in measureStacks:
            prevMeiMeasure: MeiMeasure | None = meiMeas
            meiMeas = MeiMeasure(
                measureStack,
                prevMeiMeasure,
                self,
                self.customM21AttrsToDelete,
                self.spannerBundle)
            output.append(meiMeas)

        return output

    def buildScoreElement(self, tb: TreeBuilder):
        tb.start('score', {})
        self.makeScoreDefElement(tb)

        tb.start('section', {})
        for meim in self.measures:
            meim.makeRootElement(tb)
        tb.end('section')

        tb.end('score')

    def makeScoreAndWorkElement(self, scoreNum: int) -> tuple[Element, Element]:
        tb: TreeBuilder = TreeBuilder(insert_comments=True, insert_pis=True)
        self.buildScoreElement(tb)
        scoreEl: Element = tb.close()
        tb = TreeBuilder(insert_comments=True, insert_pis=True)
        self.metadata.buildOneOfManyWorkElements(tb, scoreNum)
        workEl: Element = tb.close()
        return scoreEl, workEl

    def buildBackElement(self, tb: TreeBuilder):
        htxItems: list[MeiMetadataItem] = self.metadata.contents.get('HTX', [])
        if htxItems:
            allTheSameLanguage: bool
            theLanguage: str | None
            theLanguage, allTheSameLanguage = MeiMetadata.getTextListLanguage(htxItems)

            tb.start('back', {})
            tb.start('div', {'type': 'textTranslation'})
            if allTheSameLanguage and theLanguage:
                tb.start('lg', {'xml:lang': theLanguage})
            else:
                tb.start('lg', {})

            for htxItem in htxItems:
                # <l> can't take @analog, so use @type instead (says Perry)
                attrib: dict[str, str] = {'type': 'humdrum:HTX'}
                if isinstance(htxItem.value, m21.metadata.Text):
                    if htxItem.value.language and not allTheSameLanguage:
                        attrib['xml:lang'] = htxItem.value.language.lower()
                tb.start('l', attrib)
                tb.data(htxItem.meiValue)
                tb.end('l')
            tb.end('lg')
            tb.end('div')
            tb.end('back')

    def makeBackElement(self) -> Element:
        tb: TreeBuilder = TreeBuilder(insert_comments=True, insert_pis=True)
        self.buildBackElement(tb)
        root: Element = tb.close()
        return root

    def makeMeiElement(self) -> Element:
        tb: TreeBuilder = TreeBuilder(insert_comments=True, insert_pis=True)
        if self.meiVersion.startswith('4'):
            tb.start('mei', {
                'xmlns': 'http://www.music-encoding.org/ns/mei',
                'meiversion': '4.0.1'
            })
        elif self.meiVersion.startswith('5'):
            tb.start('mei', {
                'xmlns': 'http://www.music-encoding.org/ns/mei',
                'meiversion': '5.1+CMN'
            })
        elif not self.meiVersion:
            # empty version means we already put in in <meiCorpus>
            tb.start('mei', {})

        # meiHead
        self.metadata.buildMeiHeadElement(tb)

        if self.metadata.mainWorkXmlId:
            tb.start('music', {'decls': '#' + self.metadata.mainWorkXmlId})
        else:
            tb.start('music', {})
        tb.start('body', {})
        tb.start('mdiv', {})

        self.buildScoreElement(tb)

        tb.end('mdiv')
        tb.end('body')

        # There's one bit of metadata that goes in music/back/div@type="textTranslation":
        # humdrum:HTX
        self.buildBackElement(tb)

        tb.end('music')
        tb.end('mei')
        root: Element = tb.close()
        return root

    def makeStaffDefElement(
        self,
        part: m21.stream.Part,
        staffN: int,
        tb: TreeBuilder
    ):
        # staffLines (defaults to 5)
        staffLines: int = 5
        initialStaffLayout: m21.layout.StaffLayout | None = None
        initialStaffLayouts = list(
            part.recurse()
            .getElementsByClass(m21.layout.StaffLayout)
            .getElementsByOffsetInHierarchy(0.0)
        )
        if initialStaffLayouts:
            initialStaffLayout = initialStaffLayouts[0]
        if initialStaffLayout and initialStaffLayout.staffLines is not None:
            staffLines = initialStaffLayout.staffLines

        tb.start('staffDef', {
            'n': str(staffN),
            'lines': str(staffLines)
        })

        # clef
        clef: m21.clef.Clef | None = part.clef
        if clef is None:
            clefs: list[m21.clef.Clef] = list(
                part.recurse()
                .getElementsByClass(m21.clef.Clef)
                .getElementsByOffsetInHierarchy(0.0)
            )
            if clefs:
                clef = clefs[0]
        if clef is not None:
            M21ObjectConvert.m21ClefToMei(clef, self.spannerBundle, tb)
            clef.mei_handled_already = True  # type: ignore
            M21Utilities.extendCustomM21Attributes(
                self.customM21AttrsToDelete,
                clef,
                ['mei_handled_already']
            )
        else:
            environLocal.warn(f'No initial clef found in part {staffN}')

        # key signature
        keySig: m21.key.KeySignature | None = part.keySignature
        if keySig is None:
            keySigs: list[m21.key.KeySignature] = list(
                part.recurse()
                .getElementsByClass(m21.key.KeySignature)
                .getElementsByOffsetInHierarchy(0.0)
            )
            if keySigs:
                keySig = keySigs[0]
        if keySig is not None:
            M21ObjectConvert.m21KeySigToMei(keySig, self.spannerBundle, tb)
            keySig.mei_handled_already = True  # type: ignore
            M21Utilities.extendCustomM21Attributes(
                self.customM21AttrsToDelete,
                keySig,
                ['mei_handled_already']
            )
        else:
            environLocal.warn(f'No initial key signature found in part {staffN}')

        # time signature
        meterSig: m21.meter.TimeSignature | None = part.timeSignature
        if meterSig is None:
            meterSigs: list[m21.meter.TimeSignature] = list(
                part.recurse()
                .getElementsByClass(m21.meter.TimeSignature)
                .getElementsByOffsetInHierarchy(0.0)
            )
            if meterSigs:
                meterSig = meterSigs[0]
        if meterSig is not None:
            M21ObjectConvert.m21TimeSigToMei(meterSig, self.spannerBundle, tb)
            meterSig.mei_handled_already = True  # type: ignore
            M21Utilities.extendCustomM21Attributes(
                self.customM21AttrsToDelete,
                meterSig,
                ['mei_handled_already']
            )
        else:
            environLocal.warn(f'No initial time signature found in part {staffN}')

        tb.end('staffDef')

    _M21_STAFFGROUP_SYMBOL_TO_MEI: dict[str | None, str] = {
        # m21 symbol can be 'brace', 'bracket', 'square', 'line', None
        # mei symbol can be 'brace', 'bracket', 'bracketsq', 'line', 'none'
        'brace': 'brace',
        'bracket': 'bracket',
        'square': 'bracketsq',
        'line': 'line',
        None: 'none'
    }

    def makeStaffGrpElement(self, sgtree: M21StaffGroupTree, tb: TreeBuilder) -> list[int]:
        attr: dict[str, str] = {}
        if sgtree.staffGroup.name:
            attr['label'] = sgtree.staffGroup.name
        elif sgtree.staffGroup.abbreviation:
            attr['label'] = sgtree.staffGroup.abbreviation

        if sgtree.staffGroup.barTogether == 'Mensurstrich':
            attr['bar.method'] = 'mensur'
        elif sgtree.staffGroup.barTogether is False:
            attr['bar.thru'] = 'false'
        elif sgtree.staffGroup.barTogether is True:
            attr['bar.thru'] = 'true'

        attr['symbol'] = self._M21_STAFFGROUP_SYMBOL_TO_MEI[sgtree.staffGroup.symbol]

        tb.start('staffGrp', attr)

        part: m21.stream.Part
        staffNumsToProcess: set[int] = set(sgtree.staffNums)
        staffNums: list[int] = []
        for subgroup in sgtree.children:
            # 1. any staffNums before this subgroup
            for staffNum in sgtree.staffNums:
                if staffNum >= subgroup.lowestStaffNumber:
                    # we're done with staffNums before this subgroup
                    break

                if staffNum not in staffNumsToProcess:
                    # we already did this one
                    continue

                part = self.m21PartsForStaffNumbers[staffNum]
                self.makeStaffDefElement(part, staffNum, tb)
                staffNumsToProcess.remove(staffNum)
                staffNums.append(staffNum)

            # 2. now the subgroup
            newStaffNums: list[int] = self.makeStaffGrpElement(subgroup, tb)
            staffNumsProcessed: set[int] = set(newStaffNums)  # for speed of "in" checking
            staffNumsToProcess = set(num for num in staffNumsToProcess
                                            if num not in staffNumsProcessed)
            staffNums += newStaffNums

        # 3. any staffNums at the end after all the subgroups
        remainder: list[int] = sorted(list(staffNumsToProcess))
        for staffNum in remainder:
            part = self.m21PartsForStaffNumbers[staffNum]
            self.makeStaffDefElement(part, staffNum, tb)
            staffNumsToProcess.remove(staffNum)
            staffNums.append(staffNum)

        tb.end('staffGrp')
        staffNums.sort()
        return staffNums

    def _findTopStaffGroupTreeStartingWith(self, staffNum: int) -> M21StaffGroupTree | None:
        # self.staffGroupTrees is sorted by numStaves, so we need the last tree we saw
        # to get the topmost staffGroupTree that starts with staffNum.
        topTreeSoFar: M21StaffGroupTree | None = None
        for sgTree in self.staffGroupTrees:
            if sgTree.lowestStaffNumber == staffNum:
                topTreeSoFar = sgTree
        return topTreeSoFar

    def makeStaffGrpOrStaffDefElementStartingAt(self, staffNum: int, tb: TreeBuilder) -> int:
        # Assumption: staff numbers are in the same monotonically increasing order
        # as m21Score.parts, starting at 1, incrementing by 1 for each Part.  This
        # assumption can be seen to be true in _getStaffNumbersForM21Parts().
        sgTree: M21StaffGroupTree | None = self._findTopStaffGroupTreeStartingWith(staffNum)
        if sgTree is None:
            m21Part: m21.stream.Part = self.m21PartsForStaffNumbers[staffNum]
            self.makeStaffDefElement(m21Part, staffNum, tb)
            return staffNum + 1

        staffNums: list[int] = self.makeStaffGrpElement(sgTree, tb)
        return staffNums[-1] + 1

    def _nonGroupedStavesExist(self) -> bool:
        groupedStaffNums: set[int] = set()
        for sgTree in self.staffGroupTrees:
            groupedStaffNums.update(sgTree.staffNums)
        for sn in self.m21PartsForStaffNumbers:
            if sn not in groupedStaffNums:
                return True
        return False


    def makeScoreDefElement(self, tb: TreeBuilder) -> None:
        tb.start('scoreDef', {})
        madeFakeTopLevelStaffGrp: bool = False
        if len(self.staffGroupTrees) > 1 or self._nonGroupedStavesExist():
            # we'll need a top-level staffGrp to hold the multiple staffGroups/staffDefs
            # (just make it out of thin air)
            tb.start('staffGrp', {})
            madeFakeTopLevelStaffGrp = True

        nextStaffNum: int = 1
        while nextStaffNum in self.m21PartsForStaffNumbers:
            nextStaffNum = self.makeStaffGrpOrStaffDefElementStartingAt(nextStaffNum, tb)

        if madeFakeTopLevelStaffGrp:
            # close out the top-level staffGrp we made out of thin air
            tb.end('staffGrp')

        tb.end('scoreDef')

    def annotateScore(self) -> None:
        # self.spannerBundle has not yet been set, so make a copy of the spannerBundle
        # that was imported, for use during fillOttavas.
        spannerBundleForOttavas: m21.spanner.SpannerBundle = self.m21Score.spannerBundle

        for partIdx, part in enumerate(self.m21Score.parts):
            # We need to annotate PedalMark spanners that are in a particular
            # part (i.e. not just up in the score) with which partIdx they are
            # in, so we can later compute the correct pedal@staff value.
            for pm in part[m21.expressions.PedalMark]:
                pm.mei_part_idx = partIdx  # type: ignore
                M21Utilities.extendCustomM21Attributes(
                    self.customM21AttrsToDelete,
                    pm,
                    ['mei_part_idx']
                )

            partTieSpanners: list[tuple[M21TieSpanner, int]] = self.currentTieSpanners[part]
            for measure in part:
                if not isinstance(measure, m21.stream.Measure):
                    continue

                # check/increment numMeasuresSearched in all part tie spanners
                defunctTieSpanners: list[tuple[M21TieSpanner, int]] = []
                for i, (sp, numMeasuresSearched) in enumerate(partTieSpanners):
                    if numMeasuresSearched >= 2:
                        defunctTieSpanners.append((sp, numMeasuresSearched))
                    else:
                        partTieSpanners[i] = (sp, numMeasuresSearched + 1)
                for sp, numMeasuresSearched in defunctTieSpanners:
                    partTieSpanners.remove((sp, numMeasuresSearched))

                voices: list[m21.stream.Voice | m21.stream.Measure] = (
                    list(measure[m21.stream.Voice])
                )
                if not voices:
                    voices = [measure]
                for voice in voices:
                    voiceId: int | str
                    if isinstance(voice, m21.stream.Measure):
                        voiceId = ''
                    else:
                        voiceId = voice.id

                    for obj in voice:
                        self.annotateBeams(obj, part, voiceId)
                        self.annotateTuplets(obj, part)
                        self.annotateTies(obj, part)
                        self.annotatePositionedRests(obj, part)
                        self.fillOttavas(obj, part, spannerBundleForOttavas)

                    if self.currentTupletSpanners[part]:
                        # assume tuplets end at end of each voice
                        self.currentTupletSpanners[part] = []

                if partTieSpanners:
                    # We must have encountered a stop in a voice before we encountered the
                    # start in a different voice.  Just run through the voices again,
                    # looking for tie stops only, to pick up any we missed.
                    for voice in voices:
                        for obj in voice:
                            self.annotateTies(obj, part, stopsOnly=True)

                if self.currentTupletSpanners[part]:
                    # We have at least one missing tuplet stop.  Assume all tuplets stop
                    # at end of measure. This isn't strictly true, but iohumdrum.cpp makes
                    # the same assumption (never producing a <tupletSpan>).
                    self.currentTupletSpanners[part] = []

        # print('done annotating score')

    def annotatePositionedRests(
        self,
        obj: m21.base.Music21Object,
        part: m21.stream.Part,
    ) -> None:
        if not isinstance(obj, m21.note.Rest):
            return
        if obj.stepShift == 0:
            return

        # obj is a positioned rest, compute and stash off MEI ploc/oloc strings
        currentClef: m21.clef.PitchClef | None = obj.getContextByClass(m21.clef.PitchClef)
        if currentClef is None or not hasattr(currentClef, 'lowestLine'):
            currentClef = m21.clef.TrebleClef()
            # this should not be common enough to
            # worry about the overhead
        if t.TYPE_CHECKING:
            assert isinstance(currentClef, m21.clef.PitchClef)
        midLineDNN = currentClef.lowestLine + 4
        restObjectPseudoDNN = midLineDNN + obj.stepShift
        tempPitch = m21.pitch.Pitch()
        tempPitch.diatonicNoteNum = restObjectPseudoDNN
        obj.mei_ploc = tempPitch.step.lower()  # type: ignore
        obj.mei_oloc = str(tempPitch.octave)  # type: ignore
        M21Utilities.extendCustomM21Attributes(
            self.customM21AttrsToDelete,
            obj,
            ['mei_ploc', 'mei_oloc']
        )

    def fillOttavas(
        self,
        obj: m21.base.Music21Object,
        part: m21.stream.Part,
        spannerBundle: m21.spanner.SpannerBundle
    ) -> None:
        for spanner in obj.getSpannerSites([m21.spanner.Ottava]):
            if not M21Utilities.isIn(spanner, spannerBundle):
                continue
            if spanner.isFirst(obj):
                spanner.fill(part)

    def deannotateScore(self) -> None:
        for sp in list(self.m21Score[M21TemporarySpanner]):
            self.m21Score.remove(sp)

        if hasattr(self.m21Score, 'xml_id'):
            delattr(self.m21Score, 'xml_id')

        for obj in self.m21Score.recurse():
            # we don't put 'xml_id' in self.customM21AttrsToDelete, because
            # then every object would be in that dictionary.
            if hasattr(obj, 'xml_id'):
                delattr(obj, 'xml_id')

        for obj, customAttrs in self.customM21AttrsToDelete.items():
            for customAttr in customAttrs:
                if hasattr(obj, customAttr):
                    delattr(obj, customAttr)

        # all done, let go of these references to music21 objects.
        self.customM21AttrsToDelete = {}

#     def makeXmlIds(self) -> None:
#         for part in self.m21Score.parts:
#             for measure in part:
#                 if not isinstance(measure, m21.stream.Measure):
#                     continue
#                 self.makeMeasureXmlIds(measure, self.spannerBundle)

#     def makeMeasureXmlIds(
#         self,
#         measure: m21.stream.Measure,
#         spannerBundle: m21.spanner.SpannerBundle
#     ):
#         for obj in measure.recurse().getElementsByClass(m21.note.GeneralNote):
#             if isinstance(obj, m21.chord.Chord):
#                 # M21TieSpanner might contain notes that are inside chords (this
#                 # is not actually valid for real spanners, but it's ok for us).
#                 # Handle this as a special case.
#                 for note in obj.notes:
#                     for spanner in note.getSpannerSites():
#                         if not M21Utilities.isIn(spanner, spannerBundle):
#                             continue
#                         if isinstance(spanner, M21TieSpanner):
#                             M21Utilities.assureXmlId(spanner.getFirst())
#                             if spanner.getLast() is not spanner.getFirst():
#                                 M21Utilities.assureXmlId(spanner.getLast())
#
#             # check expressions for turn/trill/mordent; if any present, obj needs xmlId
#             for expr in obj.expressions:
#                 if isinstance(expr, (
#                     m21.expressions.Turn,
#                     m21.expressions.Trill,
#                     m21.expressions.GeneralMordent,
#                     m21.expressions.Fermata,
#                     m21.expressions.ArpeggioMark
#                 )):
#                     M21Utilities.assureXmlId(obj)
#                     break  # skip the rest of the expressions
#
#             # check for spanners (all spanners for now, might get too many xmlIds?)
#             for spanner in obj.getSpannerSites():
#                 if not M21Utilities.isIn(spanner, spannerBundle):
#                     continue
#                 if isinstance(spanner, (M21BeamSpanner, M21TupletSpanner)):
#                     # Beam spanners and tuplet spanners only need xmlIds if they
#                     # span multiple measures.  Note that we won't know this until
#                     # we encounter a later object in such a spanner, so when we
#                     # do, we reach back and assure xmlIds for everything in the
#                     # spanner.  Note that assureXmlId returns immediately if
#                     # the xml id is already in place.
#                     if not M21Utilities.allSpannedElementsAreInHierarchy(spanner, measure):
#                         for el in spanner.getSpannedElements():
#                             M21Utilities.assureXmlId(el)
#                     continue  # to next spanner
#
#                 if isinstance(spanner, M21TieSpanner):
#                     M21Utilities.assureXmlId(spanner.getFirst())
#                     if spanner.getLast() is not spanner.getFirst():
#                         M21Utilities.assureXmlId(spanner.getLast())
#                     continue  # to next spanner
#
#                 # Some spanners need xmlIds for all their elements:
#                 if isinstance(spanner, m21.expressions.ArpeggioMarkSpanner):
#                     M21Utilities.assureXmlId(obj)
#
#                 # All other spanners need xmlIds only for start and end elements
#                 if spanner.isFirst(obj) or spanner.isLast(obj):
#                     M21Utilities.assureXmlId(obj)

    def annotateBeams(
        self,
        noteOrChord: m21.base.Music21Object,
        part: m21.stream.Part,
        voiceId: int | str
    ) -> None:
        if not isinstance(noteOrChord, m21.note.NotRest):
            return

        if isinstance(noteOrChord, m21.harmony.ChordSymbol):
            # ChordSymbol is a NotRest, but should not have any beams.
            return

        def stopsBeam(beam: m21.beam.Beam) -> bool:
            if beam.type == 'stop':
                return True
            if beam.type == 'partial' and beam.direction == 'left':
                return True
            return False

        def startsBeam(beam: m21.beam.Beam) -> bool:
            if beam.type == 'start':
                return True
            if beam.type == 'partial' and beam.direction == 'right':
                return True
            return False

        def allStop(beams: m21.beam.Beams) -> bool:
            if len(beams.beamsList) == 0:
                return False
            for beamObj in beams:
                if not stopsBeam(beamObj):
                    return False
            return True

        def allStart(beams: m21.beam.Beams) -> bool:
            if len(beams.beamsList) == 0:
                return False
            for beamObj in beams:
                if not startsBeam(beamObj):
                    return False
            return True

        def computeBreakSec(prevBeams: m21.beam.Beams, currBeams: m21.beam.Beams) -> int:
            # returns the number of beams that should be seen during the break
            # returns 0 if no breaksec seen
            numBeams: int = len(currBeams)
            if len(prevBeams) != numBeams:
                return 0

            numStartStops: int = 0
            for prevBeam, currBeam in zip(prevBeams, currBeams):
                if stopsBeam(prevBeam) and startsBeam(currBeam):
                    numStartStops += 1

            if numStartStops == 0:
                return 0

            return numBeams - numStartStops

        nonIgnoredBeams: m21.beam.Beams = M21Utilities.getNonIgnoredBeams(noteOrChord)
        noteOrChord.mei_nonignored_beams = nonIgnoredBeams  # type: ignore
        M21Utilities.extendCustomM21Attributes(
            self.customM21AttrsToDelete,
            noteOrChord,
            ['mei_nonignored_beams']
        )

        # now check for sure
        if not nonIgnoredBeams.beamsList:
            self.previousBeamedNoteOrChord[(part, voiceId)] = None
            return

        if allStart(nonIgnoredBeams) or not self.currentBeamSpanners[(part, voiceId)]:
            newBeamSpanner = M21BeamSpanner()
            self.m21Score.append(newBeamSpanner)
            self.currentBeamSpanners[(part, voiceId)].append(newBeamSpanner)

        self.currentBeamSpanners[(part, voiceId)][-1].addSpannedElements(noteOrChord)

        if allStop(nonIgnoredBeams):
            # done with this <beam> or <beamSpan>.  Put the spanner in the score,
            # and clear out any state variables.
            self.currentBeamSpanners[(part, voiceId)] = (
                self.currentBeamSpanners[(part, voiceId)][:-1]
            )
            self.previousBeamedNoteOrChord[(part, voiceId)] = None
            return

        # annotate any breaksec ending at this noteOrChord (set it on the previous noteOrChord)
        prevNC = self.previousBeamedNoteOrChord[(part, voiceId)]
        if prevNC is not None:
            breakSec: int = computeBreakSec(
                prevNC.mei_nonignored_beams,  # type: ignore
                nonIgnoredBeams
            )
            if breakSec > 0:
                prevNC.mei_breaksec = breakSec  # type: ignore
                M21Utilities.extendCustomM21Attributes(
                    self.customM21AttrsToDelete,
                    prevNC,
                    ['mei_breaksec']
                )
        self.previousBeamedNoteOrChord[(part, voiceId)] = noteOrChord

    def annotateTuplets(self, gnote: m21.base.Music21Object, part: m21.stream.Part) -> None:
        if not isinstance(gnote, m21.note.GeneralNote):
            return

        if isinstance(gnote, m21.harmony.ChordSymbol):
            # ChordSymbol is a GeneralNote, but will not have any duration (i.e. no tuplets)
            # in the MEI file.
            return

        def stopsTuplet(tuplet: m21.duration.Tuplet) -> bool:
            if tuplet.type in ('stop', 'startStop'):
                return True
            return False

        def startsTuplet(tuplet: m21.duration.Tuplet) -> bool:
            if tuplet.type in ('start', 'startStop'):
                return True
            return False

        if not gnote.duration.tuplets and self.currentTupletSpanners[part]:
            # if gnote is a grace note, keep the tuplet running but skip the grace note
            if gnote.duration.isGrace:
                return

            # if gnote is NOT a grace note, then stop the tuplet
            self.currentTupletSpanners[part] = self.currentTupletSpanners[part][:-1]
            return

        # start any new tuplet spanners
        for tuplet in gnote.duration.tuplets:
            if (startsTuplet(tuplet)
                    or (tuplet.type is None and not self.currentTupletSpanners[part])):
                newTupletSpanner = M21TupletSpanner(tuplet)
                self.m21Score.append(newTupletSpanner)
                self.currentTupletSpanners[part].append(newTupletSpanner)

        # put this note in all the tuplet spanners
        for spanner in self.currentTupletSpanners[part]:
            spanner.addSpannedElements(gnote)

        # stop any old tuplet spanners
        for tuplet in gnote.duration.tuplets:
            if stopsTuplet(tuplet):
                self.currentTupletSpanners[part] = self.currentTupletSpanners[part][:-1]

    def annotateTies(
        self,
        noteOrChord: m21.base.Music21Object,
        part: m21.stream.Part,
        stopsOnly: bool = False
    ) -> None:
        if not isinstance(noteOrChord, (m21.note.Note, m21.chord.Chord)):
            # Note that we reject Unpitched and PercussionChord here, since
            # they have no pitches, so they cannot be tied.
            return

        if isinstance(noteOrChord, m21.harmony.ChordSymbol):
            # ChordSymbol is a Chord, but will not have any ties.
            return

        def startsTie(noteOrChord: m21.note.Note | m21.chord.Chord) -> bool:
            if isinstance(noteOrChord, m21.chord.Chord):
                # chord.tie should always be ignored, in music21 it actually returns
                # the first tie it finds in the chord's notes. Which is not particularly
                # useful for us here.
                return False
            if noteOrChord.tie is None:
                return False
            if noteOrChord.tie.type in ('start', 'continue'):
                return True
            return False

        def stopsTieInWhichSpanner(
            note: m21.note.Note,
            part: m21.stream.Part,
            partTieSpanners: list[tuple[M21TieSpanner, int]],
            parentChord: m21.chord.Chord | None = None
        ) -> tuple[M21TieSpanner, int] | None:
            # look at all the partTieSpanners and see if this note has the
            # same (exact) pitches as the start note of that spanner.
            for sp, numMeasuresSearched in partTieSpanners:
                startNote: m21.note.Note = sp.getFirst()
                if note is startNote:
                    continue

                # comparison via nameWithOctave is working around the music21
                # bug where a pitch with accidental=None is different from
                # a pitch with accidental='natural'.  Ugh.
                if startNote.pitch.nameWithOctave != note.pitch.nameWithOctave:
                    continue

                # to stop a tie, you have to start at or beyond the end of the tie's start note.
                startParentChord: m21.chord.Chord | None = sp.startParentChord
                minPartOffset: OffsetQL
                actualPartOffset: OffsetQL
                if startParentChord is not None:
                    minPartOffset = opFrac(
                        startParentChord.getOffsetInHierarchy(part)
                        + startParentChord.quarterLength
                    )
                else:
                    minPartOffset = opFrac(
                        startNote.getOffsetInHierarchy(part)
                        + startNote.quarterLength
                    )

                if parentChord is not None:
                    actualPartOffset = parentChord.getOffsetInHierarchy(part)
                else:
                    actualPartOffset = note.getOffsetInHierarchy(part)

                if (actualPartOffset < minPartOffset):
                    continue

                # we found a spanner whose startNote has the same pitches as
                # note, so return that spanner.  note stops the tie
                # represented by that spanner.
                return (sp, numMeasuresSearched)

            return None

        partTieSpanners: list[tuple[M21TieSpanner, int]] = self.currentTieSpanners[part]

        if not stopsOnly:
            if isinstance(noteOrChord, m21.chord.Chord):
                # A chord itself never starts a tie; perhaps one or more of the chord's
                # individual notes does.
                for note in noteOrChord.notes:
                    if startsTie(note):
                        if t.TYPE_CHECKING:
                            assert note.tie is not None
                        newTieSpanner = M21TieSpanner(note.tie, startParentChord=noteOrChord)
                        newTieSpanner.addSpannedElements(note)
                        self.m21Score.append(newTieSpanner)
                        partTieSpanners.append((newTieSpanner, 1))
            elif startsTie(noteOrChord):
                if t.TYPE_CHECKING:
                    assert noteOrChord.tie is not None
                newTieSpanner = M21TieSpanner(noteOrChord.tie, startParentChord=None)
                newTieSpanner.addSpannedElements(noteOrChord)
                self.m21Score.append(newTieSpanner)
                partTieSpanners.append((newTieSpanner, 1))

        if not partTieSpanners:
            # no tie stops to search for...
            return

        spN: tuple[M21TieSpanner, int] | None = None
        if isinstance(noteOrChord, m21.chord.Chord):
            # The chord itself does not stop a tie (M21TieSpanners always span notes);
            # perhaps one or more of the chord's individual notes does.
            for note in noteOrChord.notes:
                spN = stopsTieInWhichSpanner(note, part, partTieSpanners, noteOrChord)
                if spN is not None:
                    spN[0].addSpannedElements(note)
                    partTieSpanners.remove(spN)
        elif isinstance(noteOrChord, m21.note.Note):
            spN = stopsTieInWhichSpanner(noteOrChord, part, partTieSpanners)
            if spN is not None:
                spN[0].addSpannedElements(noteOrChord)
                partTieSpanners.remove(spN)
        else:
            raise MeiInternalError('noteOrChord is not Note or Chord')



# ------------------------------------------------------------------------------
# Name:          meiexceptions.py
# Purpose:       Exceptions that can be raised during MEI operations.
#
# Authors:       Greg Chapman <gregc@mac.com>
#
# Copyright:     (c) 2023-2025 Greg Chapman
# License:       MIT, see LICENSE
# ------------------------------------------------------------------------------
class MeiValidityError(Exception):
    '''When there is an otherwise-unspecified validity error that prevents parsing.'''
    pass

class MeiValueError(Exception):
    '''When an attribute has an invalid value.'''
    pass

class MeiAttributeError(Exception):
    '''When an element has an invalid attribute.'''
    pass

class MeiElementError(Exception):
    '''When an element itself is invalid.'''
    pass

class MeiInternalError(Exception):
    '''When an internal assumption is broken.'''
    pass

class MeiExportError(Exception):
    '''When an error occurs while converting to MEI.'''
    pass

class MeiImportError(Exception):
    '''When an error occurs while parsing MEI.'''
    pass
